import os
import os.path
import shutil
import tempfile

from mock import Mock, patch, call
import pytest

from tala.ddd.building.steps import verify
from tala.ddd.building.steps.verify import VerifyStepForGFGeneration, VerifyStepForGFRGLGeneration
from tala.model.ddd import DDD


class TestVerifyStepsForGFGeneration(object):
    def setup(self):
        self._temp_dir = tempfile.mkdtemp(prefix="TestVerify")
        self._cwd = os.getcwd()
        os.chdir(self._temp_dir)
        self._verify_step = None
        self._mock_ddd = None
        self._ddd_name = None
        self._language_codes = []
        self._mocked_auto_generator = None
        self._mocked_rasa_generator = None

    def teardown(self):
        os.chdir(self._cwd)
        shutil.rmtree(self._temp_dir)

    @pytest.mark.parametrize("step", [
        VerifyStepForGFGeneration,
        VerifyStepForGFRGLGeneration,
    ])
    def test_verify_does_not_create_build_directories(self, step):
        with self._patched_auto_generator(step) as MockAutoGenerator:
            self._given_mocked_auto_generator(MockAutoGenerator)
            self._given_ddd_name("MockDDD")
            self._given_mocked_ddd(is_rasa_enabled=False)
            self._given_language_codes(["eng"])
            self._given_file_exists("MockDDD/grammar/grammar_eng.xml")
            self._given_step_created(step)
            self._when_calling_build()
            self._then_files_have_not_been_written_by_auto_generator()

    def _patched_auto_generator(self, step):
        return patch("{}.{}._AutoGeneratorClass".format(verify.__name__, step.__name__), autospec=True)

    def _given_mocked_auto_generator(self, MockAutoGenerator):
        self._mocked_auto_generator = MockAutoGenerator.return_value

    def _given_ddd_name(self, name):
        self._ddd_name = name

    def _given_mocked_ddd(self, is_rasa_enabled):
        mock_ddd = Mock(spec=DDD)
        mock_ddd.name = self._ddd_name
        mock_ddd.is_rasa_enabled = is_rasa_enabled
        self._mock_ddd = mock_ddd

    def _given_language_codes(self, language_codes):
        self._language_codes = language_codes

    def _given_file_exists(self, path):
        self._ensure_dir_exists(os.path.dirname(path))
        with open(path, "w"):
            pass

    def _ensure_dir_exists(self, path):
        if not os.path.exists(path):
            os.makedirs(path)

    def _given_step_created(self, VerifyStep):
        ignore_warnings = True
        verbose = False
        self._verify_step = VerifyStep(
            self._mock_ddd, ignore_warnings, self._language_codes, verbose, self._ddd_name, "grammar"
        )

    def _when_calling_build(self):
        self._verify_step.build()

    def _then_files_have_not_been_written_by_auto_generator(self):
        self._mocked_auto_generator.write_to_file.assert_not_called()

    @pytest.mark.parametrize("step", [
        VerifyStepForGFGeneration,
        VerifyStepForGFRGLGeneration,
    ])
    def test_verify_generates_gf_grammars(self, step):
        with self._patched_auto_generator(step) as MockAutoGenerator:
            self._given_mocked_auto_generator(MockAutoGenerator)
            self._given_ddd_name("MockDDD")
            self._given_mocked_ddd(is_rasa_enabled=False)
            self._given_language_codes(["eng", "sv"])
            self._given_file_exists("MockDDD/grammar/grammar_eng.xml")
            self._given_step_created(step)
            self._when_calling_build()
            self._then_gf_grammars_have_been_generated_for(["eng", "sv"])

    def _then_gf_grammars_have_been_generated_for(self, expected_languages):
        expected_calls = [call(language) for language in expected_languages]
        self._mocked_auto_generator.generate.assert_has_calls(expected_calls)

    @pytest.mark.parametrize("step", [
        VerifyStepForGFGeneration,
        VerifyStepForGFRGLGeneration,
    ])
    @patch("{}.RasaGenerator".format(verify.__name__), autospec=True)
    def test_verify_does_not_create_build_rasa_directories(self, MockRASAGenerator, step):
        with self._patched_auto_generator(step):
            self._given_mocked_rasa_generator(MockRASAGenerator)
            self._given_ddd_name("MockDDD")
            self._given_mocked_ddd(is_rasa_enabled=True)
            self._given_language_codes(["eng"])
            self._given_file_exists("MockDDD/grammar/grammar_eng.xml")
            self._given_step_created(step)
            self._when_calling_build()
            self._then_rasa_files_have_not_been_written_by_rasa_generator()

    def _given_mocked_rasa_generator(self, MockRASAGenerator):
        self._mocked_rasa_generator = MockRASAGenerator.return_value

    def _then_rasa_files_have_not_been_written_by_rasa_generator(self):
        self._mocked_rasa_generator.generate_and_write_to_file.assert_not_called()

    @pytest.mark.parametrize("step", [
        VerifyStepForGFGeneration,
        VerifyStepForGFRGLGeneration,
    ])
    @patch("{}.RasaGenerator".format(verify.__name__), autospec=True)
    def test_verify_with_rasa_disabled_does_not_create_build_rasa_directories(self, MockRASAGenerator, step):
        with self._patched_auto_generator(step):
            self._given_mocked_rasa_generator(MockRASAGenerator)
            self._given_ddd_name("MockDDD")
            self._given_mocked_ddd(is_rasa_enabled=False)
            self._given_language_codes(["eng"])
            self._given_file_exists("MockDDD/grammar/grammar_eng.xml")
            self._given_step_created(step)
            self._when_calling_build()
            self._then_rasa_files_have_not_been_written_by_rasa_generator()

    @pytest.mark.parametrize("step", [
        VerifyStepForGFGeneration,
        VerifyStepForGFRGLGeneration,
    ])
    @patch("{}.RasaGenerator".format(verify.__name__), autospec=True)
    def test_verify_generates_rasa_model(self, MockRASAGenerator, step):
        with self._patched_auto_generator(step):
            self._given_mocked_rasa_generator(MockRASAGenerator)
            self._given_ddd_name("MockDDD")
            self._given_mocked_ddd(is_rasa_enabled=True)
            self._given_language_codes(["eng"])
            self._given_file_exists("MockDDD/grammar/grammar_eng.xml")
            self._given_step_created(step)
            self._when_calling_build()
            self._then_rasa_model_has_been_generated()

    def _then_rasa_model_has_been_generated(self):
        self._mocked_rasa_generator.generate.assert_called_once_with()

    @pytest.mark.parametrize("step", [
        VerifyStepForGFGeneration,
        VerifyStepForGFRGLGeneration,
    ])
    @patch("{}.RasaGenerator".format(verify.__name__), autospec=True)
    def test_verify_with_rasa_disabled_does_not_generate_rasa_model(self, MockRASAGenerator, step):
        with self._patched_auto_generator(step):
            self._given_mocked_rasa_generator(MockRASAGenerator)
            self._given_ddd_name("MockDDD")
            self._given_mocked_ddd(is_rasa_enabled=False)
            self._given_language_codes(["eng"])
            self._given_file_exists("MockDDD/grammar/grammar_eng.xml")
            self._given_step_created(step)
            self._when_calling_build()
            self._then_rasa_model_has_not_been_generated()

    def _then_rasa_model_has_not_been_generated(self):
        self._mocked_rasa_generator.generate.assert_not_called()



# flake8: noqa

import copy
import inspect
import re

from tala.model.plan import Plan
from tala.model.predicate import Predicate
from tala.model.action import Action
from tala.nl.gf.grammar_entry_types import Constants, Node
from tala.nl.gf.resource import NP as NpClass
from tala.nl.gf.resource import VP as VpClass
from tala.model.sort import CustomSort, BuiltinSort


class DddPyCompilerException(Exception):
    pass


IDENTITY = lambda x: x


class DddPyCompiler:
    def compile_ontology(self, *args, **kwargs):
        return OntologyCompiler().compile(*args, **kwargs)

    def compile_domain(self, *args, **kwargs):
        return DomainCompiler().compile(*args, **kwargs)

    def compile_grammar(self, *args, **kwargs):
        return GrammarCompiler().compile(*args, **kwargs)

    def decompile_grammar_node(self, ddd, languages, node):
        return GrammarCompiler().decompile_node(node)


class OntologyCompiler:
    def compile(self, description):
        self._ontology_name = description.__name__
        self._compile_sorts(description.sorts)
        return {
            "name": self._ontology_name,
            "sorts": set(self._sorts_dict.values()),
            "predicates": self._compile_predicates(description.predicates),
            "individuals": self._compile_individuals(description.individuals),
            "actions": description.actions,
        }

    def _compile_sorts(self, sorts_description):
        self._sorts_dict = {}
        for name in sorts_description:
            self._sorts_dict[name] = CustomSort(self._ontology_name, name, **sorts_description[name])

    def _compile_predicates(self, predicates_description):
        predicates = set([])
        for name, description in predicates_description.iteritems():
            predicate = self._compile_predicate(name, description)
            predicates.add(predicate)
        return predicates

    def _compile_predicate(self, name, description):
        if isinstance(description, str):
            sort_name = description
            return Predicate(self._ontology_name, name, self._get_sort(sort_name))
        else:
            description_copy = copy.copy(description)
            sort_name = description_copy.pop("sort")
            sort = self._get_sort(sort_name)
            if "feature_of" in description_copy:
                feature_of_name = description_copy.pop("feature_of")
            else:
                feature_of_name = None
            return Predicate(self._ontology_name, name, sort=sort, feature_of_name=feature_of_name, **description_copy)

    def _get_sort(self, name):
        if name in self._sorts_dict:
            return self._sorts_dict[name]
        else:
            return self._get_builtin_sort(name)

    def _get_builtin_sort(self, name):
        return BuiltinSort(name)

    def _compile_individuals(self, individuals_description):
        individuals = {}
        for individual_name in individuals_description:
            sort_name = individuals_description[individual_name]
            sort = self._get_sort(sort_name)
            individuals[individual_name] = sort
        return individuals


class DomainCompiler:
    def compile(self, ddd_name, description, ontology, parser):
        self._ddd_name = ddd_name
        self._domain_description = description
        self._domain_name = self._domain_description.__name__
        self._ontology = ontology
        self._parser = parser
        self._parse_plans()
        self._parse_default_questions()
        self._parse_parameters()
        self._parse_dependencies()
        return {
            "ddd_name": self._ddd_name,
            "name": self._domain_name,
            "plans": self._plans,
            "default_questions": self._default_questions,
            "parameters": self._parameters,
            "dependencies": self._dependencies
        }

    def get_name(self, description):
        return description.__name__

    def _parse_plans(self):
        if not isinstance(self._domain_description.plans, list):
            raise DddPyCompilerException("plans should be a list")
        self._plans = []
        for plan_description in self._domain_description.plans:
            parsed_plan = self._parse_plan(plan_description)
            goal = parsed_plan["goal"]
            self._plans.append(parsed_plan)

    def _parse_plan(self, plan_description):
        self._plan_description = copy.copy(plan_description)
        plan = self._parse_plan_from_description()
        if len(self._plan_description) > 0:
            raise DddPyCompilerException("unsupported plan descriptor(s): %s" % self._plan_description)
        return plan

    def _parse_plan_from_description(self):
        plan = {}
        plan["goal"] = self._compile_goal(self._select_plan_descriptor("goal"))
        plan["plan"] = self._compile_plan(self._select_plan_descriptor("plan"))
        self._compile_plan_single_attribute(plan, "preferred", self._parse_boolean)
        self._compile_plan_single_attribute(plan, "accommodate_without_feedback", self._parse_boolean)
        self._compile_plan_single_attribute(plan, "dynamic_title")
        self._compile_plan_single_attribute(plan, "postplan", self._compile_plan)
        self._compile_plan_multi_attribute(plan, "superactions", self._compile_superaction)
        self._compile_plan_multi_attribute(plan, "gui_context", self._compile_gui_context)
        self._compile_plan_multi_attribute(plan, "postconds", self._parse)
        self._compile_plan_single_attribute(plan, "io_status")
        self._compile_plan_single_attribute(plan, "restart_on_completion")
        return plan

    def _compile_goal(self, string):
        potential_goal = self._parse(string)
        if potential_goal.is_goal():
            return potential_goal
        else:
            if potential_goal.is_action():
                raise DddPyCompilerException(
                    'in %s: "%s" is not a goal. Perhaps you mean "perform(%s)"?' %
                    (inspect.getsourcefile(self._domain_description), string, string)
                )
            elif potential_goal.is_question():
                raise DddPyCompilerException(
                    'in %s: "%s" is not a goal. Perhaps you mean "resolve(%s)"?' %
                    (inspect.getsourcefile(self._domain_description), string, string)
                )
            else:
                raise DddPyCompilerException(
                    "in %s: Expected goal but found %s (parsed as %s)" %
                    (inspect.getsourcefile(self._domain_description), unicode(string), unicode(potential_goal))
                )

    def _compile_plan_single_attribute(self, plan, attribute_name, compilation_method=IDENTITY):
        if attribute_name in self._plan_description:
            attribute = self._select_plan_descriptor(attribute_name)
            plan[attribute_name] = compilation_method(attribute)

    def _compile_plan_multi_attribute(self, plan, attribute_name, compilation_method=IDENTITY):
        if attribute_name in self._plan_description:
            plan[attribute_name] = map(
                lambda string: compilation_method(string), self._select_plan_descriptor(attribute_name)
            )

    def _compile_superaction(self, string):
        return Action(string, self._ontology.get_name())

    def _compile_gui_context(self, predicate_name):
        return self._ontology.get_predicate(predicate_name)

    def _select_plan_descriptor(self, key):
        value = self._plan_description[key]
        del self._plan_description[key]
        return value

    def _parse_boolean(self, value):
        if value == True:
            return True
        else:
            return self._parse(value)

    def _compile_plan(self, plan_items_as_strings):
        if not isinstance(plan_items_as_strings, list):
            raise DddPyCompilerException("plan not list: %r" % plan_items_as_strings)
        items = [self._parse(string) for string in plan_items_as_strings]
        return Plan(reversed(items))

    def _parse_default_questions(self):
        if hasattr(self._domain_description, "default_questions"):
            self._default_questions = map(self._parse, self._domain_description.default_questions)
        else:
            self._default_questions = []

    def _parse_parameters(self):
        if hasattr(self._domain_description, "parameters"):
            self._parameters = {}
            for question_as_string in self._domain_description.parameters.keys():
                question = self._parse(question_as_string)
                parameters_as_string = self._domain_description.parameters[question_as_string]
                parameters = self._parser.parse_parameters(parameters_as_string)
                self._parameters[question] = parameters
        else:
            self._parameters = {}

    def _parse_dependencies(self):
        self._dependencies = {}
        if hasattr(self._domain_description, "dependencies"):
            for dependent_string in self._domain_description.dependencies:
                dependent = self._parse(dependent_string)
                others = set()
                for other_string in self._domain_description.dependencies[dependent_string]:
                    others.add(self._parse(other_string))
                self._dependencies[dependent] = others

    def _parse(self, string):
        return self._parser.parse(string)


class KeyParser:
    def __init__(self, entry_type, pattern, *conditions):
        self._entry_type = entry_type
        self._pattern = pattern
        self._re = re.compile(self._pattern_to_regexp(pattern))
        self._conditions = conditions

    def _pattern_to_regexp(self, pattern):
        return pattern.replace("$", "(\w+)")

    def parse(self, string):
        match = self._re.match(string)
        if match:
            args = match.groups()
            if self._conditions_apply(args):
                return self._tuple_with_entry_type_and_args(args)

    def _tuple_with_entry_type_and_args(self, args):
        return tuple([self._entry_type] + list(args))

    def _conditions_apply(self, args):
        for arg, condition in zip(args, self._conditions):
            if not condition(arg):
                return False
        return True

    def generate(self, args_dict):
        return self._replace_placeholders_in_pattern_with_args(args_dict)

    def _replace_placeholders_in_pattern_with_args(self, args_dict):
        result = self._pattern
        if self._entry_type in GrammarCompiler.PARAMETER_MAP:
            parameter_names = GrammarCompiler.PARAMETER_MAP[self._entry_type]
            for parameter_name in parameter_names:
                value = args_dict[parameter_name]
                result = result.replace("$", value, 1)
        return result


class GrammarCompiler:
    PARAMETER_MAP = {
        Constants.ACTION_TITLE: ("action", ),
        Constants.ACTION: ("name", ),
        Constants.SYS_ANSWER: ("predicate", ),
        Constants.USER_QUESTION: ("predicate", ),
        Constants.INDIVIDUAL: ("name", ),
        Constants.ISSUE_TITLE: ("predicate", ),
        Constants.VALIDITY: ("name", ),
        Constants.SYS_QUESTION: ("predicate", ),
        Constants.REPORT_STARTED: ("action", ),
        Constants.REPORT_FAILED: ("action", "reason"),
        Constants.REPORT_ENDED: ("action", ),
        Constants.PREREPORT: ("action", ),
        Constants.PREDICATE: ("name", ),
        Constants.PRECONFIRM: ("action", ),
        Constants.POSITIVE_SYS_ANSWER: ("predicate", ),
        Constants.NEGATIVE_SYS_ANSWER: ("predicate", )
    }

    def __init__(self):
        self._key_parsers = {}
        self._add_key_parser(Constants.USER_QUESTION, "$_user_question", self._is_predicate)
        self._add_key_parser(Constants.SYS_QUESTION, "$_sys_question", self._is_predicate)
        self._add_key_parser(Constants.PREDICATE, "$", self._is_predicate)
        self._add_key_parser(Constants.ACTION, "$", self._is_action)
        self._add_key_parser(Constants.ANSWER_COMBINATION, "ANSWER")
        self._add_key_parser(Constants.PREREPORT, "$_prereport", self._is_service_action)
        self._add_key_parser(Constants.PRECONFIRM, "$_preconfirm", self._is_service_action)
        self._add_key_parser(Constants.REPORT_STARTED, "$_started", self._is_service_action)
        self._add_key_parser(Constants.REPORT_ENDED, "$_ended", self._is_service_action)
        self._add_key_parser(Constants.REPORT_FAILED, "$_failed_$", self._is_service_action)
        self._add_key_parser(Constants.INDIVIDUAL, "$", self._is_individual)
        self._add_key_parser(Constants.SYS_ANSWER, "$_sys_answer", self._is_predicate)
        self._add_key_parser(Constants.POSITIVE_SYS_ANSWER, "$_positive_sys_answer", self._is_predicate)
        self._add_key_parser(Constants.NEGATIVE_SYS_ANSWER, "$_negative_sys_answer", self._is_predicate)
        self._add_key_parser(Constants.ACTION_TITLE, "$_title", self._is_action)
        self._add_key_parser(Constants.ISSUE_TITLE, "$_title", self._is_predicate)
        self._add_key_parser(Constants.VALIDITY, "$", self._is_validity)
        self._add_key_parser(Constants.GREETING, "greeting")

    def _add_key_parser(self, entry_type, pattern, *conditions):
        self._key_parsers[entry_type] = KeyParser(entry_type, pattern, *conditions)

    def _is_predicate(self, string):
        return self._ontology.has_predicate(string)

    def _is_action(self, string):
        return self._ontology.is_action(string)

    def _is_individual(self, string):
        return string in self._ontology.get_individuals()

    def _is_validity(self, string):
        if self._service_interface:
            for validator in self._service_interface.validators:
                if validator.name == string:
                    return True

    def _is_service_action(self, string):
        if self._service_interface:
            return self._service_interface.has_action(string)

    def compile(self, string, ontology, service_interface):
        self._ontology = ontology
        self._service_interface = service_interface
        module_dict = dict()
        exec (string, module_dict)
        return self._compile_entries_from_dict(module_dict)

    def _compile_entries_from_dict(self, module_dict):
        grammar_node = Node(Constants.GRAMMAR, {})
        for name, entry in module_dict.iteritems():
            if not self._is_builtin_member(name) and not self._is_resource_class(name):
                self._compile_entry(grammar_node, name, entry)
        return grammar_node

    def _is_builtin_member(self, name):
        return name.startswith("__")

    def _is_resource_class(self, name):
        return name in ["NP", "VP", "MASCULINE", "FEMININE", "SINGULAR", "PLURAL"]

    def _compile_entry(self, parent_node, key_string, form_object):
        key = self._parse_key(key_string)
        if key:
            node_type = key[0]
            form_node = self._compile_form(form_object)
            parameters = self._get_parameters(key, form_object)
            entry_node = Node(node_type, parameters, [form_node])
            parent_node.add_child(entry_node)
        else:
            self._warn("failed to parse grammar entry key %r" % key_string)

    def _get_parameters(self, key, form_object):
        parameters = self._get_parameters_via_map(key)
        node_specific_parameters = self._get_node_specific_parameters(key, form_object)
        if node_specific_parameters is not None:
            parameters.update(node_specific_parameters)
        return parameters

    def _get_parameters_via_map(self, key):
        node_type = key[0]
        parameters = {}
        if len(key) > 1:
            parameter_names = self.PARAMETER_MAP[node_type]
            for name, value in zip(parameter_names, key[1:]):
                parameters[name] = value
        return parameters

    def _get_node_specific_parameters(self, key, form_object):
        node_type = key[0]
        if node_type == Constants.ACTION:
            if isinstance(form_object, list):
                return self._get_action_specific_parameters_from_form_objects(form_object)
            elif isinstance(form_object, NpClass):
                return self._get_action_specific_parameters_from_form_object(form_object)

    def _get_action_specific_parameters_from_form_objects(self, form_objects):
        for form_object in form_objects:
            parameters = self._get_action_specific_parameters_from_form_object(form_object)
            if parameters is not None:
                return parameters

    def _get_action_specific_parameters_from_form_object(self, form_object):
        if isinstance(form_object, NpClass):
            result = {}
            if form_object.gender is not None:
                result["gender"] = form_object.gender
            if form_object.number is not None:
                result["number"] = form_object.number
            return result

    def _parse_key(self, string):
        for parser in self._key_parsers.values():
            key = parser.parse(string)
            if key:
                return key

    def _compile_form(self, form_object):
        if isinstance(form_object, list):
            return self._compile_one_of(form_object)
        elif self._is_of_resource_class(form_object):
            return self._resource_class_node(form_object)
        elif isinstance(form_object, (str, unicode)):
            return self._compile_form_string(form_object)
        else:
            raise DddPyCompilerException("failed to compile grammar form %r" % form_object)

    def _compile_one_of(self, obj):
        result = Node(Constants.ONE_OF, {})
        for child_object in obj:
            compiled_child = self._compile_form(child_object)
            if compiled_child.type == Constants.ITEM:
                result.add_child(compiled_child)
            else:
                result.add_child(Node(Constants.ITEM, {}, [compiled_child]))
        return result

    def _is_of_resource_class(self, obj):
        return self._is_resource_class(obj.__class__.__name__)

    def _resource_class_node(self, resource_object):
        if isinstance(resource_object, NpClass):
            return self._create_np_node(resource_object)
        if isinstance(resource_object, VpClass):
            return self._create_vp_node(resource_object)

    def _create_np_node(self, resource_np):
        children = []
        children.append(Node(Constants.INDEFINITE, {}, [resource_np.indefinite]))
        if resource_np.definite is not None:
            children.append(Node(Constants.DEFINITE, {}, [resource_np.definite]))
        np_node = Node(Constants.NP, {}, children)
        return np_node

    def _create_vp_node(self, resource_vp):
        infinitive_node = Node(Constants.INFINITIVE, {}, [resource_vp.infinitive])
        imperative_node = Node(Constants.IMPERATIVE, {}, [resource_vp.imperative])
        ing_form_node = Node(Constants.ING_FORM, {}, [resource_vp.ing_form])
        object_node = Node(Constants.OBJECT, {}, [resource_vp.object])
        vp_node = Node(Constants.VP, {}, [infinitive_node, imperative_node, ing_form_node, object_node])
        return vp_node

    def _compile_form_string(self, string):
        parts = re.split("(<answer:[^<>]+>|<individual>)", string)
        child_nodes = [self._compile_form_part(part) for part in parts if part != ""]
        return Node(Constants.ITEM, {}, child_nodes)

    def _compile_form_part(self, string):
        if string == "<individual>":
            return Node(Constants.SLOT)
        else:
            match = re.match("^<answer:([^<>]+)>$", string)
            if match:
                predicate_or_sort = match.group(1)
                if self._ontology.has_predicate(predicate_or_sort):
                    return Node(Constants.SLOT, {"predicate": predicate_or_sort})
                elif self._ontology.has_sort(predicate_or_sort):
                    return Node(Constants.SLOT, {"sort": predicate_or_sort})
                else:
                    raise DddPyCompilerException("expected predicate or sort but found %r" % predicate_or_sort)
            else:
                return string

    def decompile_node(self, node):
        key_string = self._decompile_key(node)
        value_string = self._decompile_node_value(node)
        return '%s = %s\n' % (key_string, value_string)

    def _decompile_key(self, key):
        return self._key_parsers[key.type].generate(key.parameters)

    def _decompile_node_value(self, node):
        if isinstance(node, (str, unicode)):
            return node
        elif node.type == Constants.SLOT:
            return "<answer:%s>" % node.parameters["predicate"]
        elif node.type == Constants.ONE_OF:
            result = "[\n"
            for child in node.children:
                result += "  %s,\n" % self._decompile_node_value(child)
            result += "]"
            return result
        else:
            children_string = "".join([self._decompile_node_value(child) for child in node.children])
            if self._children_is_a_one_of(node.children):
                return children_string
            else:
                return '"%s"' % children_string

    def _children_is_a_one_of(self, children):
        return len(children) == 1 and isinstance(children[0], Node) and children[0].type == Constants.ONE_OF

    def _warn(self, warning):
        print "warning: %s" % warning



from __future__ import print_function

import os
import shutil

from tala.ddd.building.steps.abstract_build_step import AbstractBuildStep
from tala.nl.languages import SUPPORTED_RASA_LANGUAGES
from tala.utils import chdir


class AbstractCleanStep(AbstractBuildStep):
    _build_directory = None

    def __init__(self, ddd, ignore_warnings, language_codes, verbose, ddd_root_directory, grammar_directory):
        super(AbstractCleanStep, self).__init__(ddd, language_codes, ddd_root_directory, grammar_directory, verbose)
        self._ignore_warnings = ignore_warnings

    def build(self):
        with chdir.chdir(self._ddd_root_directory):
            with chdir.chdir(self._grammar_directory):
                self._create_empty_build_directories()
                self._potentially_create_empty_rasa_build_directories()

    def _create_empty_build_directories(self):
        for language_code in self._language_codes:
            self._create_empty_build_directory(language_code, self._build_directory)

    def _potentially_create_empty_rasa_build_directories(self):
        if not self._ddd.is_rasa_enabled:
            print("Not using RASA NLU, will not clean RASA build directories.")
            return

        for language_code in self._language_codes:
            self._create_empty_rasa_build_directory(language_code)

    def _create_empty_rasa_build_directory(self, language_code):
        if language_code in SUPPORTED_RASA_LANGUAGES:
            directory = self._directory_of_rasa_files(language_code)
            print("[%s] Cleaning RASA build directory '%s'..." % (language_code, directory), end="")
            self._clean_directory(directory)
            print("Done.")

    def _create_empty_build_directory(self, language_code, directory_name):
        directory = "%s/%s" % (directory_name, language_code)
        print("[%s] Cleaning build directory '%s'..." % (language_code, directory), end="")
        self._clean_directory(directory)
        print("Done.")

    def _clean_directory(self, directory):
        if os.path.exists(directory):
            shutil.rmtree(directory, ignore_errors=True)
        os.makedirs(directory)


class CleanStepForGFGeneration(AbstractCleanStep):
    _build_directory = "build"


class CleanStepForHandcraftedGFFiles(AbstractCleanStep):
    _build_directory = "build_handcrafted"



from __future__ import print_function

from tala.ddd.building.steps.abstract_build_step import AbstractBuildStep
from tala.nl.gf.auto_generator import AutoGenerator
from tala.nl.gf.rgl_gf_generator import RglGfFilesGenerator
from tala.nl.rasa.generating.generator import RasaGenerator
from tala.utils import chdir


class AbstractVerifyStep(AbstractBuildStep):
    _id = None
    _build_directory = None
    _AutoGeneratorClass = None

    def __init__(self, ddd, ignore_warnings, language_codes, verbose, ddd_root_directory, grammar_directory):
        super(AbstractVerifyStep, self).__init__(ddd, language_codes, ddd_root_directory, grammar_directory, verbose)
        self._ignore_warnings = ignore_warnings

    def build(self):
        with chdir.chdir(self._ddd_root_directory):
            print("Verifying models for DDD '%s'." % self._name)
            with chdir.chdir(self._grammar_directory):
                for language_code in self._language_codes:
                    self._verify_grammars(language_code)
                    self._potentially_verify_rasa_model(language_code)
            print("Finished verifying models for DDD '%s'." % self._name)

    def _potentially_verify_rasa_model(self, language_code):
        if not self._ddd.is_rasa_enabled:
            print("[%s] RASA NLU is disabled, skipping its grammar verification." % (language_code))
            return

        self._verify_rasa_nlu_model(language_code)

    def _verify_rasa_nlu_model(self, language_code):
        print("[%s] Verifying grammar for RASA NLU." % (language_code))
        generator = RasaGenerator(self._ddd, language_code)
        generator.generate()
        print("[%s] Finished verifying grammar for RASA NLU." % (language_code))

    def _verify_grammars(self, language_code):
        raise NotImplementedError()

    def _verify_gf_grammar(self, AutoGeneratorClass, language_code):
        generator = AutoGeneratorClass(self._ddd, self._ignore_warnings, cwd=self._ddd_root_directory)
        generator.generate(language_code)
        generator.clear()


class VerifyStepForGFGeneration(AbstractVerifyStep):
    _build_directory = "build"
    _id = "generated"
    _AutoGeneratorClass = AutoGenerator

    def _verify_grammars(self, language_code):
        print("[%s] Verifying grammar." % (language_code))
        self._verify_gf_grammar(self._AutoGeneratorClass, language_code)
        print("[%s] Finished verifying grammar." % (language_code))


class VerifyStepForHandcraftedGFFiles(AbstractVerifyStep):
    _build_directory = "build_handcrafted"
    _id = "handcrafted"

    def _verify_grammars(self, language_code):
        print("[%s] Using handcrafted grammar, will not verify." % (language_code))
        pass


class VerifyStepForGFRGLGeneration(VerifyStepForGFGeneration):
    _AutoGeneratorClass = RglGfFilesGenerator



# flake8: noqa

from tala.model.move import ICMMove, ICMMoveWithSemanticContent
from tala.model.plan_item import RespondPlanItem, BindPlanItem, ConsultDBPlanItem, IfThenElse, ServiceReportPlanItem, DoPlanItem, PlanItem, QuitPlanItem, GreetPlanItem, EmitIcmPlanItem, ForgetAllPlanItem, ForgetPlanItem, ForgetIssuePlanItem, InvokeServiceQueryPlanItem, InvokeServiceActionPlanItem, JumpToPlanItem, AssumePlanItem, AssumeSharedPlanItem, AssumeIssuePlanItem, MinResultsNotSupportedException, MaxResultsNotSupportedException
from tala.model.question_raising_plan_item import FindoutPlanItem, RaisePlanItem
from tala.model.proposition import ServiceResultProposition
from tala.model.test.utils import LibTestCase


class PlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.question = self.dest_city_question
        self.respond_plan_item = RespondPlanItem(self.question)
        self.findout_plan_item = FindoutPlanItem(self.domain_name, self.question)
        self.raise_plan_item = RaisePlanItem(self.domain_name, self.question)
        self.bind_plan_item = BindPlanItem(self.question)
        self.consult_db_plan_item = ConsultDBPlanItem(self.question)
        self.if_condition = self.proposition_dest_city_paris
        self.if_consequent = self.findout_plan_item
        self.if_alternative = self.raise_plan_item
        self.if_then_else_plan_item = IfThenElse(self.if_condition, self.if_consequent, self.if_alternative)
        self.service_result_proposition = ServiceResultProposition(
            self.ontology_name, "mockup_service_action", [], "mock_perform_result"
        )
        self.service_report_item = ServiceReportPlanItem(self.service_result_proposition)

    def test_equality_with_none(self):
        self.assertNotEqual(self.respond_plan_item, None)

    def test_create_DoPlanItem(self):
        do_item = DoPlanItem(self.buy_action)
        self.assertEquals(self.buy_action, do_item.getContent())

    def test_DoPlanItem_type(self):
        do_item = DoPlanItem(self.buy_action)
        self.assertEquals(PlanItem.TYPE_DO, do_item.getType())

    def test_isRespondPlanItem(self):
        self.assertTrue(self.respond_plan_item.isRespondPlanItem())

    def test_isFindoutPlanItem(self):
        self.assertTrue(self.findout_plan_item.isFindoutPlanItem())

    def test_FindoutPlanItem_getContent(self):
        self.assertEquals(self.question, self.findout_plan_item.getContent())

    def test_isRaisePlanItem(self):
        self.assertTrue(self.raise_plan_item.isRaisePlanItem())

    def test_RaisePlanItem_getContent(self):
        self.assertEquals(self.question, self.raise_plan_item.getContent())

    def test_FindoutPlanItem_to_string(self):
        expected_string = "findout(?X.dest_city(X))"
        self.assertEquals(expected_string, unicode(self.findout_plan_item))

    def test_PlanItem_unicode(self):
        quitPlanItem = QuitPlanItem()
        self.assertEquals(unicode(quitPlanItem), "quit")

    def test_GreetPlanItem_unicode(self):
        greetPlanItem = GreetPlanItem()
        self.assertEquals(unicode(greetPlanItem), "greet")

    def test_findout_is_question_plan_item(self):
        self.assertTrue(self.findout_plan_item.is_question_plan_item())

    def test_raise_is_question_plan_item(self):
        self.assertTrue(self.raise_plan_item.is_question_plan_item())

    def test_bind_is_question_plan_item(self):
        self.assertTrue(self.bind_plan_item.is_question_plan_item())

    def test_consultDB_is_not_question_plan_item(self):
        self.assertFalse(self.consult_db_plan_item.is_question_plan_item())

    def test_respond_is_not_question_plan_item(self):
        self.assertFalse(self.respond_plan_item.is_question_plan_item())

    def test_findout_is_question_raising_item(self):
        self.assertTrue(self.findout_plan_item.is_question_raising_item())

    def test_raise_is_question_raising_item(self):
        self.assertTrue(self.raise_plan_item.is_question_raising_item())

    def test_bind_is_not_question_raising_item(self):
        self.assertFalse(self.bind_plan_item.is_question_raising_item())

    def test_consultDB_is_not_question_raising_item(self):
        self.assertFalse(self.consult_db_plan_item.is_question_raising_item())

    def test_respond_is_not_question_raising_item(self):
        self.assertFalse(self.respond_plan_item.is_question_raising_item())

    def test_emit_item_for_icm_und_non_pos_is_question_raising_item(self):
        icm_und = ICMMove(ICMMove.UND, polarity=ICMMove.NEG)
        emit_item = EmitIcmPlanItem(icm_und)
        self.assertTrue(emit_item.is_question_raising_item())

    def test_emit_item_for_icm_und_pos_for_positive_content_is_question_raising_item(self):
        positive_content = self.proposition_dest_city_paris
        icm = ICMMoveWithSemanticContent(ICMMove.UND, positive_content, polarity=ICMMove.POS)
        emit_item = EmitIcmPlanItem(icm)
        self.assertTrue(emit_item.is_question_raising_item())

    def test_emit_item_for_icm_und_pos_for_negated_content_is_not_question_raising_item(self):
        negation = self.proposition_not_dest_city_paris
        icm = ICMMoveWithSemanticContent(ICMMove.UND, negation, polarity=ICMMove.POS)
        emit_item = EmitIcmPlanItem(icm)
        self.assertFalse(emit_item.is_question_raising_item())

    def test_report_item_is_turn_yielding(self):
        self.assertTrue(self.service_report_item.is_turn_yielding())

    def test_respond_item_is_turn_yielding(self):
        self.assertTrue(self.respond_plan_item.is_turn_yielding())

    def test_emit_item_for_acc_neg_is_turn_yielding(self):
        icm_acc_neg = ICMMove(ICMMove.ACC, polarity=ICMMove.NEG)
        emit_item = EmitIcmPlanItem(icm_acc_neg)
        self.assertTrue(emit_item.is_turn_yielding())

    def test_emit_item_for_acc_pos_is_not_turn_yielding(self):
        icm_acc_pos = ICMMove(ICMMove.ACC, polarity=ICMMove.POS)
        emit_item = EmitIcmPlanItem(icm_acc_pos)
        self.assertFalse(emit_item.is_turn_yielding())

    def test_if_then_else_getters(self):
        plan_item = self.if_then_else_plan_item
        self.assertEquals(PlanItem.TYPE_IF_THEN_ELSE, plan_item.getType())
        self.assertEquals(self.if_condition, plan_item.get_condition())
        self.assertEquals(self.if_consequent, plan_item.get_consequent())
        self.assertEquals(self.if_alternative, plan_item.get_alternative())

    def test_if_then_else_to_string(self):
        expected_string = "if_then_else(%s, %s, %s)" % (self.if_condition, self.if_consequent, self.if_alternative)
        self.assertEquals(expected_string, unicode(self.if_then_else_plan_item))

    def test_create_forget_all(self):
        item = ForgetAllPlanItem()
        self.assertTrue(item.is_forget_all_plan_item())

    def test_create_forget_plan_item_proposition(self):
        fact = self.proposition_dest_city_paris
        item = ForgetPlanItem(fact)
        self.assertTrue(item.is_forget_plan_item())
        self.assertEqual(fact, item.getContent())

    def test_create_forget_plan_item_predicate(self):
        predicate = self.predicate_dept_city
        item = ForgetPlanItem(predicate)
        self.assertTrue(item.is_forget_plan_item())
        self.assertEqual(predicate, item.getContent())

    def test_create_forget_issue_plan_item(self):
        issue = self.price_question
        item = ForgetIssuePlanItem(issue)
        self.assertTrue(item.is_forget_issue_plan_item())
        self.assertEqual(issue, item.getContent())

    def test_service_report_plan_item(self):
        self.assertTrue(self.service_report_item.is_service_report_plan_item())
        self.assertTrue(self.service_result_proposition, self.service_report_item.getContent())

    def test_findout_equality(self):
        item = FindoutPlanItem(self.domain_name, self.question)
        identical_item = FindoutPlanItem(self.domain_name, self.question)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(identical_item, item)

    def test_inequality_due_to_question(self):
        item1 = FindoutPlanItem(self.domain_name, self.dest_city_question)
        item2 = FindoutPlanItem(self.domain_name, self.price_question)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(item1, item2)

    def test_clone_findout_into_raise_plan_item(self):
        findout_plan_item = FindoutPlanItem(self.domain_name, self.question)
        raise_plan_item = findout_plan_item.clone_as_type(PlanItem.TYPE_RAISE)
        self.assertTrue(raise_plan_item.isRaisePlanItem())
        self.assertEquals(self.question, raise_plan_item.get_question())


class InvokeServiceActionPlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_is_invoke_service_action_plan_item(self):
        self.given_created_service_action_plan_item()
        self.when_call(self._plan_item.is_invoke_service_action_plan_item)
        self.then_result_is(True)

    def given_created_service_action_plan_item(self, service_action="mock_service_action", **kwargs):
        self._plan_item = InvokeServiceActionPlanItem(self.ontology.name, service_action, **kwargs)

    def test_get_service_action(self):
        self.given_created_service_action_plan_item(service_action="mock_service_action")
        self.when_call(self._plan_item.get_service_action)
        self.then_result_is("mock_service_action")

    def test_has_interrogative_preconfirmation_false_by_default(self):
        self.given_created_service_action_plan_item()
        self.when_call(self._plan_item.has_interrogative_preconfirmation)
        self.then_result_is(False)

    def test_interrogative_preconfirmation_true(self):
        self.given_created_service_action_plan_item(preconfirm=InvokeServiceActionPlanItem.INTERROGATIVE)
        self.when_call(self._plan_item.has_interrogative_preconfirmation)
        self.then_result_is(True)

    def test_has_assertive_preconfirmation_false_by_default(self):
        self.given_created_service_action_plan_item()
        self.when_call(self._plan_item.has_assertive_preconfirmation)
        self.then_result_is(False)

    def test_assertive_preconfirmation_true(self):
        self.given_created_service_action_plan_item(preconfirm=InvokeServiceActionPlanItem.ASSERTIVE)
        self.when_call(self._plan_item.has_assertive_preconfirmation)
        self.then_result_is(True)

    def test_has_postconfirmation_false_by_default(self):
        self.given_created_service_action_plan_item()
        self.when_call(self._plan_item.has_postconfirmation)
        self.then_result_is(False)

    def test_has_postconfirmation_true(self):
        self.given_created_service_action_plan_item(postconfirm=True)
        self.when_call(self._plan_item.has_postconfirmation)
        self.then_result_is(True)

    def test_should_downdate_plan_true_by_default(self):
        self.given_created_service_action_plan_item()
        self.when_call(self._plan_item.should_downdate_plan)
        self.then_result_is(True)

    def test_should_downdate_plan_false(self):
        self.given_created_service_action_plan_item(downdate_plan=False)
        self.when_call(self._plan_item.should_downdate_plan)
        self.then_result_is(False)

    def test_unicode(self):
        self.given_created_service_action_plan_item(
            service_action="mock_service_action", preconfirm=None, postconfirm=False, downdate_plan=True
        )
        self.when_get_unicode()
        self.then_result_is(
            "invoke_service_action(mock_service_action, {preconfirm=None, postconfirm=False, downdate_plan=True})"
        )

    def when_get_unicode(self):
        self._actual_result = unicode(self._plan_item)


class InvokeServiceQueryPlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def test_is_invoke_service_query_plan_item(self):
        self.given_created_invoke_service_query_plan_item()
        self.when_call(self._plan_item.is_invoke_service_query_plan_item)
        self.then_result_is(True)

    def given_created_invoke_service_query_plan_item(self, *args, **kwargs):
        self._create_invoke_service_query_plan_item(*args, **kwargs)

    def _create_invoke_service_query_plan_item(self, issue="mock_issue", min_results=None, max_results=None):
        self._plan_item = InvokeServiceQueryPlanItem(issue, min_results, max_results)

    def test_min_results_is_0_by_default(self):
        self.given_created_invoke_service_query_plan_item()
        self.when_call(self._plan_item.get_min_results)
        self.then_result_is(0)

    def test_min_results_overridden(self):
        self.given_created_invoke_service_query_plan_item(min_results=1)
        self.when_call(self._plan_item.get_min_results)
        self.then_result_is(1)

    def test_max_results_is_none_by_default(self):
        self.given_created_invoke_service_query_plan_item()
        self.when_call(self._plan_item.get_max_results)
        self.then_result_is(None)

    def test_max_results_overridden(self):
        self.given_created_invoke_service_query_plan_item(max_results=1)
        self.when_call(self._plan_item.get_max_results)
        self.then_result_is(1)

    def test_exception_raised_for_min_results_below_0(self):
        self.when_created_invoke_service_query_plan_item_then_exception_is_raised(
            min_results=-1,
            max_results=None,
            expected_exception=MinResultsNotSupportedException,
            expected_message="Expected 'min_results' to be 0 or above but got -1."
        )

    def test_exception_raised_for_max_results_below_1(self):
        self.when_created_invoke_service_query_plan_item_then_exception_is_raised(
            min_results=0,
            max_results=0,
            expected_exception=MaxResultsNotSupportedException,
            expected_message="Expected 'max_results' to be None or above 0 but got 0."
        )

    def when_created_invoke_service_query_plan_item_then_exception_is_raised(
        self, min_results, max_results, expected_exception, expected_message
    ):
        with self.assertRaises(expected_exception) as context_manager:
            self._create_invoke_service_query_plan_item(min_results=min_results, max_results=max_results)
        self.assertEquals(expected_message, str(context_manager.exception))

    def test_getContent(self):
        self.given_created_invoke_service_query_plan_item(issue="mock_issue")
        self.when_call(self._plan_item.getContent)
        self.then_result_is("mock_issue")

    def test_unicode(self):
        self.given_created_invoke_service_query_plan_item(issue="mock_issue", min_results=0, max_results=1)
        self.when_get_unicode()
        self.then_result_is("invoke_service_query(mock_issue, min_results=0, max_results=1)")

    def when_get_unicode(self):
        self._actual_result = unicode(self._plan_item)


class consultDBPlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()

    def testConsultDBPlanItemCreation(self):
        plan_item = ConsultDBPlanItem(self.price_question)
        self.assertTrue(plan_item.isConsultDBPlanItem())

    def testConsultDBPlanItemContent(self):
        plan_item = ConsultDBPlanItem(self.price_question)
        self.assertEquals(self.price_question, plan_item.getContent())


class JumpToPlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.action = self.buy_action
        self.item = JumpToPlanItem(self.action)

    def test_unicode(self):
        self.assertEquals("jumpto(buy)", unicode(self.item))

    def test_equality(self):
        identical_item = JumpToPlanItem(self.action)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(self.item, identical_item)

    def test_inequality(self):
        non_identical_item = JumpToPlanItem(self.top_action)
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.item, non_identical_item)


class AssumePlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.assume_price = AssumePlanItem(self.price_proposition)

    def test_unicode(self):
        self.assertEquals("assume(price(1234.0))", unicode(self.assume_price))

    def test_is_assume_item(self):
        self.assertTrue(self.assume_price.is_assume_plan_item())


class AssumeSharedPlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.assume_shared_price = AssumeSharedPlanItem(self.price_proposition)

    def test_unicode(self):
        self.assertEquals("assume_shared(price(1234.0))", unicode(self.assume_shared_price))

    def test_is_assume_shared_item(self):
        self.assertTrue(self.assume_shared_price.is_assume_shared_plan_item())


class AssumeSharedIssuePlanItemTests(LibTestCase):
    def setUp(self):
        self.setUpLibTestCase()
        self.assume_price_issue = AssumeIssuePlanItem(self.price_question)

    def test_unicode(self):
        self.assertEquals("assume_issue(?X.price(X))", unicode(self.assume_price_issue))

    def test_is_assume_shared_issue(self):
        self.assertTrue(self.assume_price_issue.is_assume_issue_plan_item())



class TooFewTextChunksException(Exception):
    pass


class InvalidTextChunkEntityCombinationException(Exception):
    pass


class Intent(object):
    def __init__(self, intent, text_chunks, required_entities):
        self._intent = intent
        self._text_chunks = text_chunks
        self._required_entities = required_entities

        if len(text_chunks) < 1:
            raise TooFewTextChunksException("Expected at least one text chunk but got %s" % text_chunks)

        if len(text_chunks) != len(required_entities) + 1:
            raise InvalidTextChunkEntityCombinationException(
                "Expected %d text chunks for the %d entities, to achieve [chunk, entity, chunk, [...], entity, chunk] "
                "but got %d chunks" % (len(required_entities) + 1, len(required_entities), len(text_chunks))
            )

    @property
    def intent(self):
        return self._intent

    @property
    def text_chunks(self):
        return self._text_chunks

    @property
    def required_entities(self):
        return self._required_entities

    def __eq__(self, other):
        return bool(
            isinstance(other, self.__class__) and self._intent == other.intent
            and self._text_chunks == other.text_chunks and self._required_entities == other.required_entities
        )

    def __repr__(self):
        return "%s(%r, text_chunks=%s, required_entities=%s)" % (
            self.__class__.__name__, self._intent, self._text_chunks, self._required_entities
        )


class Request(Intent):
    def __init__(self, action, text_chunks, required_entities):
        super(Request, self).__init__(action, text_chunks, required_entities)

    @property
    def action(self):
        return self._intent


class Question(Intent):
    def __init__(self, predicate, text_chunks, required_entities):
        super(Question, self).__init__(predicate, text_chunks, required_entities)

    @property
    def predicate(self):
        return self._intent


class Answer(Intent):
    def __init__(self, text_chunks, required_entities):
        super(Answer, self).__init__("answer", text_chunks, required_entities)



import tala.nl.gf.resource
from tala.nl.gf.resource import VP, SINGULAR, PLURAL, MASCULINE, FEMININE  # noqa: F401


class NP(tala.nl.gf.resource.NP):
    def __init__(self, indefinite, gender, number=SINGULAR):
        self.indefinite = indefinite
        self.number = number
        self.gender = gender
        self.definite = None


top = [NP("avvia visualizzazione", "xxxxxxxxxxx"), "cancella", "ricomincia", "stop"]

up = ["indietro", "torna indietro"]



import copy

from tala.model.domain import Domain
from tala.model.goal import ResolveGoal, PerformGoal, HandleGoal
from tala.model.lambda_abstraction import LambdaAbstractedPredicateProposition
from tala.model.speaker import Speaker
from tala.model.set import Set
from tala.model.ontology import Ontology
from tala.model.plan import Plan, InvalidPlansException
from tala.model.plan_item import BindPlanItem, IfThenElse, InvokeServiceActionPlanItem
from tala.model.polarity import Polarity
from tala.model.proposition import PredicateProposition, GoalProposition, PropositionSet, ServiceActionTerminatedProposition
from tala.model.question import WhQuestion, AltQuestion
from tala.model.question_raising_plan_item import QuestionRaisingPlanItem, FindoutPlanItem, RaisePlanItem
from tala.model.sort import CustomSort, RealSort
from tala.model.test.utils import LibTestCase


class DomainTests(LibTestCase):
    def setUp(self):
        self.ontology_name = "mockup_ontology"
        self._city_sort = CustomSort(self.ontology_name, "city", dynamic=True)
        self.setUpDomainTestCase()

    def setUpDomainTestCase(self):
        self._create_ontologies()
        self._create_semantic_objects()
        self._create_domains()

    def _create_ontologies(self):
        sorts = {
            CustomSort(self.ontology_name, "country"),
            self._city_sort,
            CustomSort(self.ontology_name, "city_type"),
        }
        predicates = {
            self._create_predicate("dest_city", self._city_sort),
            self._create_predicate("dept_city", self._city_sort),
            self._create_predicate("dest_country", CustomSort(self.ontology_name, "country")),
            self._create_predicate("price", RealSort()),
            self._create_predicate("code", RealSort()),
            self._create_predicate(
                "dest_city_type", sort=CustomSort(self.ontology_name, "city_type"), feature_of_name="dest_city"
            )
        }
        individuals = {
            "paris": self._city_sort,
            "london": self._city_sort,
        }
        self.actions = {
            "top",
            "buy",
            "always_preferred",
            "conditionally_preferred",
            "planless",
            "with_context",
            "with_dynamic_title",
            "accommodate_unrestricted",
            "downdate_plan_true",
            "downdate_plan_false",
            "event",
        }
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, self.actions)

        self.empty_ontology = Ontology("empty_ontology", {}, {}, {}, set([]))

    def _create_semantic_objects(self):
        self.predicate_dest_city = self.ontology.get_predicate("dest_city")
        self.predicate_dest_city_type = self.ontology.get_predicate("dest_city_type")
        self.predicate_dept_city = self.ontology.get_predicate("dept_city")
        self.predicate_price = self.ontology.get_predicate("price")
        self.predicate_dest_country = self.ontology.get_predicate("dest_country")
        self.predicate_code = self.ontology.get_predicate("code")

        self.lambda_abstracted_price_prop = LambdaAbstractedPredicateProposition(
            self.predicate_price, self.ontology_name
        )
        self.lambda_abstracted_dest_city_prop = LambdaAbstractedPredicateProposition(
            self.predicate_dest_city, self.ontology_name
        )
        self.lambda_abstracted_dest_city_type_prop = LambdaAbstractedPredicateProposition(
            self.predicate_dest_city_type, self.ontology_name
        )
        self.lambda_abstracted_dept_city_prop = LambdaAbstractedPredicateProposition(
            self.predicate_dept_city, self.ontology_name
        )

        self.price_question = WhQuestion(self.lambda_abstracted_price_prop)
        self.dest_city_question = WhQuestion(self.lambda_abstracted_dest_city_prop)
        self.dest_city_type_question = WhQuestion(self.lambda_abstracted_dest_city_type_prop)
        self.dept_city_question = WhQuestion(self.lambda_abstracted_dept_city_prop)
        self.dest_country_question = WhQuestion(
            LambdaAbstractedPredicateProposition(self.predicate_dest_country, self.ontology_name)
        )

        self.individual_paris = self.ontology.create_individual("paris")
        self.individual_london = self.ontology.create_individual("london")
        self.proposition_dest_city_paris = PredicateProposition(self.predicate_dest_city, self.individual_paris)
        self.proposition_dest_city_london = PredicateProposition(self.predicate_dest_city, self.individual_london)
        self.proposition_not_dest_city_paris = PredicateProposition(
            self.predicate_dest_city, self.individual_paris, Polarity.NEG
        )

    def _create_domains(self):
        self.domain_name = "mockup_domain"

        self._create_actions()
        self._create_questions()
        self.condition = self.proposition_dest_city_paris
        plans = self._create_plans()

        dependencies = {self.dest_country_question: {self.dest_city_question}}

        self.non_verbalized_question = self.dest_city_question

        parameters = {
            self.non_verbalized_question: {
                "verbalize": False
            },
        }

        self.domain = Domain(
            self.DDD_NAME,
            self.domain_name,
            self.ontology,
            plans=plans,
            default_questions=[self.dest_city_question],
            dependencies=dependencies,
            parameters=parameters
        )

        self.empty_domain = Domain(self.DDD_NAME, "empty_domain", self.empty_ontology)

    def _create_actions(self):
        for action_name in self.actions:
            action = self.ontology.create_action(action_name)
            attribute_name = action_name + "_action"
            setattr(self, attribute_name, action)
        self.up_action = self.ontology.create_action("up")

    def _create_questions(self):
        for predicate_name in ["dest_city", "dept_city", "code"]:
            question = self.ontology.create_wh_question(predicate_name)
            attribute_name = predicate_name + "_question"
            setattr(self, attribute_name, question)

    def _create_plans(self):
        return [{
            "goal":
            ResolveGoal(self.price_question, Speaker.SYS),
            "plan":
            Plan([
                FindoutPlanItem(self.domain_name, self.dest_city_question),
                FindoutPlanItem(self.domain_name, self.dept_city_question),
            ]),
            "gui_context": [
                self.predicate_dest_city,
                self.predicate_dept_city,
            ]
        },
                {
                    "goal": PerformGoal(self.buy_action),
                    "plan": Plan([FindoutPlanItem(self.domain_name, self.price_question)]),
                    "postconds": [self.condition]
                },
                {
                    "goal": PerformGoal(self.always_preferred_action),
                    "preferred": True,
                    "plan": Plan([]),
                    "postplan": [FindoutPlanItem(self.domain_name, self.price_question)],
                    "superactions": [self.buy_action]
                },
                {
                    "goal": PerformGoal(self.conditionally_preferred_action),
                    "preferred": self.condition,
                    "plan": Plan([])
                },
                {
                    "goal": PerformGoal(self.with_context_action),
                    "plan": Plan([]),
                    "gui_context": [self.predicate_dest_city]
                },
                {
                    "goal": PerformGoal(self.with_dynamic_title_action),
                    "plan": Plan([]),
                    "dynamic_title": True,
                    "gui_context": [self.predicate_dest_city]
                },
                {
                    "goal": PerformGoal(self.accommodate_unrestricted_action),
                    "plan": Plan([]),
                    "unrestricted_accommodation": True,
                },
                {
                    "goal":
                    PerformGoal(self.downdate_plan_true_action),
                    "plan":
                    Plan([InvokeServiceActionPlanItem("mockup_ontology", "mock_service_action", downdate_plan=True)]),
                },
                {
                    "goal":
                    PerformGoal(self.downdate_plan_false_action),
                    "plan":
                    Plan([InvokeServiceActionPlanItem("mockup_ontology", "mock_service_action", downdate_plan=False)]),
                }, {
                    "goal": HandleGoal("mockup_ontology", "event"),
                    "plan": Plan([]),
                }]

    def test_top_and_up_goals_added_implicitly_by_domain(self):
        self.given_custom_goals_without_top_and_up()
        self.when_fetching_goals_from_domain()
        self.then_top_and_up_are_included()

    def given_custom_goals_without_top_and_up(self):
        plans = self._create_plans()
        goals = [info["goal"] for info in plans]
        self.expected_goals = goals
        action_names = self._action_names_of_perform_goals(self.expected_goals)
        self.assertNotIn("top", action_names)
        self.assertNotIn("up", action_names)

    def when_fetching_goals_from_domain(self):
        self.actual_goals = self.domain.goals

    def then_top_and_up_are_included(self):
        self.assertEqual(len(self.actual_goals), len(self.expected_goals) + 2)
        action_names = self._action_names_of_perform_goals(self.actual_goals)
        self.assertIn("top", action_names)
        self.assertIn("up", action_names)

    def _action_names_of_perform_goals(self, goals):
        actions_of_perform_goals = [goal.get_content() for goal in goals if goal.is_perform_goal()]
        action_names = [action.value for action in actions_of_perform_goals]
        return action_names

    def test_domain_equality(self):
        identical_domain = copy.copy(self.domain)
        self.assert_eq_returns_true_and_ne_returns_false_symmetrically(self.domain, identical_domain)

    def test_domain_inequality(self):
        non_identical_domain = self.empty_domain
        self.assert_eq_returns_false_and_ne_returns_true_symmetrically(self.domain, non_identical_domain)

    def test_get_name(self):
        self.assertEquals("mockup_domain", self.domain.get_name())

    def test_action_method(self):
        self.assertEquals(self.buy_action, self.domain.action("buy"))

    def test_has_goal_true_for_question_with_plan(self):
        self.assertTrue(self.domain.has_goal(ResolveGoal(self.price_question, Speaker.SYS)))

    def test_has_goal_false_for_question_without_plan(self):
        question_without_plan = self.dest_city_question
        self.assertFalse(self.domain.has_goal(ResolveGoal(question_without_plan, Speaker.USR)))

    def test_has_goal_true_for_action_with_plan(self):
        self.assertTrue(self.domain.has_goal(PerformGoal(self.buy_action)))

    def test_has_goal_false_for_planless_action(self):
        self.assertFalse(self.domain.has_goal(PerformGoal(self.planless_action)))

    def test_up_plan_empty(self):
        self.assertEquals(0, len(self.domain.get_plan(PerformGoal(self.up_action))))

    def test_goal_is_preferred_true_if_plan_always_preferred(self):
        self.assertTrue(self.domain.goal_is_preferred(PerformGoal(self.always_preferred_action)))

    def test_goal_is_preferred_false_if_plan_never_preferred(self):
        never_preferred_action = self.buy_action
        self.assertFalse(self.domain.goal_is_preferred(PerformGoal(never_preferred_action)))

    def test_goal_is_preferred_true_if_condition_in_facts(self):
        facts_with_condition = [self.condition]
        self.assertTrue(
            self.domain.goal_is_conditionally_preferred(
                PerformGoal(self.conditionally_preferred_action), facts_with_condition
            )
        )

    def test_goal_is_preferred_false_if_condition_not_in_facts(self):
        facts_without_condition = []
        self.assertFalse(
            self.domain.goal_is_conditionally_preferred(
                PerformGoal(self.conditionally_preferred_action), facts_without_condition
            )
        )

    def test_is_default_question(self):
        default_question = self.dest_city_question
        non_default_question = self.price_question
        self.assertTrue(self.domain.is_default_question(default_question))
        self.assertFalse(self.domain.is_default_question(non_default_question))

    def test_plan_can_be_silently_accommodated(self):
        silently_accommodatable_goal = PerformGoal(self.accommodate_unrestricted_action)
        self.assertTrue(self.domain.is_silently_accommodatable(silently_accommodatable_goal))

    def test_plan_with_conditionally_preferred_action_can_not_be_silently_accommodated(self):
        silently_accommodatable_goal = PerformGoal(self.conditionally_preferred_action)
        self.assertFalse(self.domain.is_silently_accommodatable(silently_accommodatable_goal))

    def test_get_plan_questions(self):
        self.maxDiff = None
        expected_questions = ["?X.dest_city(X)", "?X.dest_city_type(X)", "?X.dept_city(X)", "?X.price(X)"]
        actual_result = [question for question in self.domain.get_plan_questions()]
        actual_result_strings = map(str, actual_result)
        self.assertItemsEqual(expected_questions, actual_result_strings)

    def test_get_all_goals(self):
        expected_goals = {
            ResolveGoal(self.price_question, Speaker.SYS),
            PerformGoal(self.buy_action),
            PerformGoal(self.always_preferred_action),
            PerformGoal(self.conditionally_preferred_action),
            PerformGoal(self.with_context_action),
            PerformGoal(self.up_action),
            PerformGoal(self.with_dynamic_title_action),
            PerformGoal(self.accommodate_unrestricted_action),
            PerformGoal(self.top_action),
            PerformGoal(self.downdate_plan_true_action),
            PerformGoal(self.downdate_plan_false_action),
            HandleGoal("mockup_ontology", "event")
        }
        actual_result = set(self.domain.get_all_goals())
        self.assertEqual(expected_goals, actual_result)

    def test_get_all_goals_in_defined_order(self):
        expected_goals = [
            ResolveGoal(self.price_question, Speaker.SYS),
            PerformGoal(self.buy_action),
            PerformGoal(self.always_preferred_action),
            PerformGoal(self.conditionally_preferred_action),
            PerformGoal(self.with_context_action),
            PerformGoal(self.with_dynamic_title_action),
            PerformGoal(self.accommodate_unrestricted_action),
            PerformGoal(self.downdate_plan_true_action),
            PerformGoal(self.downdate_plan_false_action),
            HandleGoal("mockup_ontology", "event")
        ]
        actual_result = self.domain.get_all_goals_in_defined_order()
        self.assertEqual(expected_goals, actual_result)

    def test_get_dependent_question_returns_goal_issue_for_question_in_plan(self):
        self.assertEquals(self.price_question, self.domain.get_dependent_question(self.dest_city_question))

    def test_get_dependent_question_returns_none_for_goal_issue(self):
        self.assertEquals(None, self.domain.get_dependent_question(self.price_question))

    def test_iterate_plans(self):
        expected_goals = {
            ResolveGoal(self.price_question, Speaker.SYS),
            PerformGoal(self.buy_action),
            PerformGoal(self.always_preferred_action),
            PerformGoal(self.conditionally_preferred_action),
            PerformGoal(self.with_context_action),
            PerformGoal(self.up_action),
            PerformGoal(self.with_dynamic_title_action),
            PerformGoal(self.accommodate_unrestricted_action),
            PerformGoal(self.top_action),
            PerformGoal(self.downdate_plan_true_action),
            PerformGoal(self.downdate_plan_false_action),
            HandleGoal("mockup_ontology", "event")
        }
        actual_goals = set([goal for goal in self.domain.get_plan_goal_iterator()])
        self.assertEqual(expected_goals, actual_goals)

    def test_get_postconds_for_goal_with_postcond(self):
        action_with_postconds = self.buy_action
        self.assertEqual([self.condition], list(self.domain.get_postconds(PerformGoal(action_with_postconds))))

    def test_get_postconds_for_goal_without_postconds(self):
        action_without_postcond = self.always_preferred_action
        self.assertEqual([], list(self.domain.get_postconds(PerformGoal(action_without_postcond))))

    def test_get_postconds_returns_implicit_postconds_for_service_action_invocations_if_downdate_plan_is_true(self):
        self.assertEquals([ServiceActionTerminatedProposition("mockup_ontology", "mock_service_action")],
                          list(self.domain.get_postconds(PerformGoal(self.downdate_plan_true_action))))

    def test_get_postconds_doesnt_return_implicit_postconds_for_service_action_invocations_if_not_downdating_plan(self):
        self.assertEquals([], list(self.domain.get_postconds(PerformGoal(self.downdate_plan_false_action))))

    def test_get_postconds_considers_downdate_plan_property_for_handle_goals(self):
        self.assertEquals([ServiceActionTerminatedProposition("mockup_ontology", "event")],
                          list(self.domain.get_postconds(HandleGoal("mockup_ontology", "event"))))

    def test_get_postconds_for_resolve_goal(self):
        resolve_goal = ResolveGoal(self.price_question, Speaker.SYS)
        self.assertEqual([], list(self.domain.get_postconds(resolve_goal)))

    def test_get_postplan_for_goal_with_postplan(self):
        action_with_postplan = self.always_preferred_action
        postplan = self.domain.get_postplan(PerformGoal(action_with_postplan))
        expected_postplan = [FindoutPlanItem(self.domain_name, self.price_question)]
        self.assertEqual(expected_postplan, postplan)

    def test_get_postplan_for_goal_without_postplan(self):
        action_without_postplan = self.buy_action
        self.assertEqual([], self.domain.get_postplan(PerformGoal(action_without_postplan)))

    def test_get_superactions_for_goal_with_superactions(self):
        action_with_superactions = self.always_preferred_action
        expected_result = [self.buy_action]
        self.assertEqual(expected_result, self.domain.get_superactions(PerformGoal(action_with_superactions)))

    def test_get_superactions_for_goal_without_superactions(self):
        action_without_superactions = self.buy_action
        self.assertEqual([], self.domain.get_superactions(PerformGoal(action_without_superactions)))

    def test_is_depending_on_true_for_goal_issue_wrt_question_in_its_plan(self):
        goal_issue = self.price_question
        question_in_plan = self.dest_city_question
        self.assertTrue(self.domain.is_depending_on(goal_issue, question_in_plan))

    def test_is_depending_on_false_for_plan_question_wrt_its_goal_issue(self):
        goal_issue = self.price_question
        question_in_plan = self.dest_city_question
        self.assertFalse(self.domain.is_depending_on(question_in_plan, goal_issue))

    def test_is_depending_on_true_due_to_domain_declaration(self):
        self.assertTrue(self.domain.is_depending_on(self.dest_country_question, self.dest_city_question))

    def test_is_depending_on_false_for_inverse_domain_declaration(self):
        self.assertFalse(self.domain.is_depending_on(self.dest_city_question, self.dest_country_question))

    def test_is_depending_on_true_because_of_featurehood(self):
        self.assertTrue(self.domain.is_depending_on(self.dest_city_question, self.dest_city_type_question))

    def test_is_depending_on_robust_to_non_sortal_questions(self):
        city_prop_set = PropositionSet([
            self.proposition_dest_city_london,
            self.proposition_dest_city_paris,
        ])
        non_sortal_question = AltQuestion(city_prop_set)

        self.domain.is_depending_on(non_sortal_question, non_sortal_question)

    def test_get_gui_context_for_issue_goal_with_context(self):
        expected_result = [self.predicate_dest_city, self.predicate_dept_city]
        self.assertEquals(expected_result, self.domain.get_gui_context(ResolveGoal(self.price_question, Speaker.SYS)))

    def test_goal_with_dynamic_title_has_dynamic_title(self):
        self.assertTrue(self.domain.has_dynamic_title(PerformGoal(self.with_dynamic_title_action)))

    def test_goal_without_dynamic_title_has_no_dynamic_title(self):
        self.assertFalse(self.domain.has_dynamic_title(PerformGoal(self.with_context_action)))

    def test_get_gui_context_for_action_goal_with_context(self):
        expected_result = [self.predicate_dest_city]
        self.assertEquals(expected_result, self.domain.get_gui_context(PerformGoal(self.with_context_action)))

    def test_get_gui_context_for_goal_without_context(self):
        expected_result = []
        goal_without_context = self.buy_action
        self.assertEquals(expected_result, self.domain.get_gui_context(PerformGoal(goal_without_context)))

    def test_get_resolving_answers(self):
        question = self.dest_city_question
        expected_answers = [self.proposition_dest_city_paris, self.proposition_dest_city_london]
        self.assertEqual(set(expected_answers), set(self.domain.get_resolving_answers(question)))

    def test_question_presented_as_list_forces_graphical_choice(self):
        self.domain.parameters[self.dest_city_question] = {
            "graphical_type": QuestionRaisingPlanItem.GRAPHICAL_TYPE_LIST
        }
        self.assertTrue(self.domain.question_forces_graphical_choice(self.dest_city_question))

    def test_question_presented_as_text_forces_graphical_choice(self):
        self.domain.parameters[self.dest_city_question] = {
            "graphical_type": QuestionRaisingPlanItem.GRAPHICAL_TYPE_TEXT
        }
        self.assertTrue(self.domain.question_forces_graphical_choice(self.dest_city_question))

    def test_incremental_question_does_not_force_graphical_choice(self):
        self.domain.parameters[self.dest_city_question] = {"incremental": True}
        self.assertFalse(self.domain.question_forces_graphical_choice(self.dest_city_question))

    def test_get_question_verbalize_returns_true_by_default(self):
        question_with_verbalize_unspecified = self.price_question
        self.assertTrue(self.domain.get_verbalize(question_with_verbalize_unspecified))

    def test_get_question_verbalize_returns_false_when_specified_as_false(self):
        self.assertFalse(self.domain.get_verbalize(self.non_verbalized_question))

    def test_empty_top_plan_by_default(self):
        self.assertEquals(Plan([]), self.domain.get_plan(PerformGoal(self.top_action)))

    def test_multiple_plans_for_same_goal_yields_exception(self):
        invalid_plans = [
            {
                "goal": PerformGoal(self.top_action),
                "plan": Plan([])
            },
            {
                "goal": PerformGoal(self.top_action),
                "plan": Plan([])
            },
        ]

        with self.assertRaises(InvalidPlansException):
            Domain(self.DDD_NAME, self.domain_name, self.ontology, invalid_plans)

    def test_get_questions_in_plan_base_case(self):
        plan = Plan([
            FindoutPlanItem(self.domain_name, self.price_question),
            RaisePlanItem(self.domain_name, self.dest_city_question)
        ])
        expected_questions = [self.price_question, self.dest_city_question, self.dest_city_type_question]
        actual_result = [question for question in self.domain.get_questions_in_plan(plan)]
        self.assertEqual(set(expected_questions), set(actual_result))

    def test_get_questions_in_plan_supports_if_then_else(self):
        arbitrary_condition = self.proposition_dest_city_paris
        plan = Plan([
            IfThenElse(
                arbitrary_condition, FindoutPlanItem(self.domain_name, self.price_question),
                BindPlanItem(self.dest_city_question)
            )
        ])
        expected_questions = [self.price_question, self.dest_city_question, self.dest_city_type_question]
        actual_result = [question for question in self.domain.get_questions_in_plan(plan)]
        self.assertEqual(set(expected_questions), set(actual_result))

    def test_get_questions_in_plan_include_features(self):
        plan = Plan([FindoutPlanItem(self.domain_name, self.dest_city_question)])
        expected_questions = [self.dest_city_question, self.dest_city_type_question]
        actual_result = [question for question in self.domain.get_questions_in_plan(plan)]
        self.assertEqual(set(expected_questions), set(actual_result))


class DominatesTests(LibTestCase):
    def setUp(self):
        self.ontology_name = "mockup_ontology"
        actions = {
            "dominating_action", "dominated_action", "super_dominated_action", "non_dominated_action", "planless_action"
        }
        sorts = set([])
        predicates = {
            self._create_predicate("dominated_issue", RealSort()),
            self._create_predicate("non_dominated_issue", RealSort())
        }
        individuals = {}
        self.ontology_name = "mockup_ontology"
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

        self.domain_name = "mockup_domain"
        self.DDD_NAME = "mock_ddd"

        for action_name in actions:
            action = self.ontology.create_action(action_name)
            setattr(self, action_name, action)

        for question_name in ["dominated_issue", "non_dominated_issue"]:
            question = self.ontology.create_wh_question(question_name)
            setattr(self, question_name, question)

        plans = [{
            "goal":
            PerformGoal(self.dominating_action),
            "plan": [
                self._findout_with_alts([
                    GoalProposition(PerformGoal(self.dominated_action)),
                    GoalProposition(ResolveGoal(self.dominated_issue, Speaker.SYS))
                ])
            ]
        },
                 {
                     "goal": PerformGoal(self.dominated_action),
                     "plan": [self._findout_with_alts([GoalProposition(PerformGoal(self.super_dominated_action))])]
                 }, {
                     "goal": PerformGoal(self.super_dominated_action),
                     "plan": []
                 }, {
                     "goal": ResolveGoal(self.dominated_issue, Speaker.SYS),
                     "plan": []
                 }, {
                     "goal": PerformGoal(self.non_dominated_action),
                     "plan": []
                 }]

        self.domain = Domain(self.DDD_NAME, self.domain_name, self.ontology, plans=plans)

    def test_dominates_true_for_dominated_action(self):
        self.assertTrue(self.domain.dominates(PerformGoal(self.dominating_action), PerformGoal(self.dominated_action)))

    def test_dominates_true_for_super_dominated_action(self):
        self.assertTrue(
            self.domain.dominates(PerformGoal(self.dominating_action), PerformGoal(self.super_dominated_action))
        )

    def test_dominates_true_for_dominated_issue(self):
        self.assertTrue(
            self.domain.dominates(PerformGoal(self.dominating_action), ResolveGoal(self.dominated_issue, Speaker.SYS))
        )

    def test_dominates_false_for_non_dominated_action(self):
        self.assertFalse(
            self.domain.dominates(PerformGoal(self.dominating_action), PerformGoal(self.non_dominated_action))
        )

    def test_dominates_false_for_non_dominated_issue(self):
        self.assertFalse(
            self.domain.dominates(
                PerformGoal(self.dominating_action), ResolveGoal(self.non_dominated_issue, Speaker.SYS)
            )
        )

    def test_dominates_false_for_action_without_plan(self):
        self.assertFalse(self.domain.dominates(PerformGoal(self.planless_action), PerformGoal(self.dominated_action)))

    def _findout_with_alts(self, alts):
        return FindoutPlanItem(self.domain_name, AltQuestion(PropositionSet(alts)))


class IOStatusTests(LibTestCase):
    def setUp(self):
        self.DDD_NAME = "mockup_ddd"
        self.statuses = ["default", "excluded", "hidden", "silent", "disabled"]
        self.facts = Set()
        self._create_ontology()
        self._create_domain()

    def _create_ontology(self):
        self.ontology_name = "status_ontology"
        sorts = set([])
        predicates = set([])
        individuals = {}
        actions = set(self.statuses)
        self.ontology = Ontology(self.ontology_name, sorts, predicates, individuals, actions)

    def _create_domain(self):
        self.domain_name = "status_domain"

        for status in self.statuses:
            action = self.ontology.create_action(status)
            attribute_name = status + "_action"
            setattr(self, attribute_name, action)

        plans = [{
            "goal": PerformGoal(self.default_action),
            "plan": []
        }, {
            "goal": PerformGoal(self.excluded_action),
            "io_status": "excluded",
            "plan": []
        }, {
            "goal": PerformGoal(self.hidden_action),
            "io_status": "hidden",
            "plan": []
        }, {
            "goal": PerformGoal(self.silent_action),
            "io_status": "silent",
            "plan": []
        }, {
            "goal": PerformGoal(self.disabled_action),
            "io_status": "disabled",
            "plan": []
        }]

        self.domain = Domain(self.DDD_NAME, self.domain_name, self.ontology, plans=plans)

    def test_io_status_for_default_plan(self):
        self.assertEquals(Domain.DEFAULT_IO_STATUS, self.domain.get_io_status(PerformGoal(self.default_action)))

    def test_io_status_for_excluded_plan(self):
        self.assertEquals(Domain.EXCLUDED_IO_STATUS, self.domain.get_io_status(PerformGoal(self.excluded_action)))

    def test_io_status_for_hidden_plan(self):
        self.assertEquals(Domain.HIDDEN_IO_STATUS, self.domain.get_io_status(PerformGoal(self.hidden_action)))

    def test_io_status_for_silent_plan(self):
        self.assertEquals(Domain.SILENT_IO_STATUS, self.domain.get_io_status(PerformGoal(self.silent_action)))

    def test_io_status_for_disabled_plan(self):
        self.assertEquals(Domain.DISABLED_IO_STATUS, self.domain.get_io_status(PerformGoal(self.disabled_action)))



ACTION_INTENT = "action"
QUESTION_INTENT = "question"
ANSWER_INTENT = "answer"
ANSWER_NEGATION_INTENT = "answer_negation"
VALID_INTENT_MOVES = [ACTION_INTENT, QUESTION_INTENT]
VALID_ANSWER_MOVES = [ANSWER_INTENT, ANSWER_NEGATION_INTENT]
NEGATIVE_INTENT = "NEGATIVE"



#!/usr/bin/env python2.7

import argparse
import contextlib
import os
import re
import warnings

from tala.backend.dependencies.for_generating import BackendDependenciesForGenerating
from tala.cli.argument_parser import add_common_backend_arguments
from tala.config import BackendConfig, DddConfig, RasaConfig, BackendConfigNotFoundException, DddConfigNotFoundException, RasaConfigNotFoundException
from tala.ddd.building.ddd_builder_for_generating import DDDBuilderForGenerating
from tala import installed_version
from tala.cli import console_formatting
from tala.nl import languages
from tala.ddd.maker import utils as ddd_maker_utils
from tala.ddd.maker.ddd_maker import DddMaker


class ConfigAlreadyExistsException(Exception):
    pass


class ConfigNotFoundException(Exception):
    pass


class InvalidArgumentException(Exception):
    pass


def create_ddd(args):
    def directory_to_class_name(directory_name):
        name_with_capitalized_words = directory_name.title()
        class_name = re.sub("[_ ]", "", name_with_capitalized_words)
        return class_name

    directory_name = args.name
    class_name = directory_to_class_name(directory_name)
    DddMaker(class_name, directory_name, args.target_dir).make()


def create_backend_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            "Expected to be able to create backend config file '%s' but it already exists." % args.filename
        )
    BackendConfig().write_default_config(args.filename, args.ddd)


def create_ddd_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            "Expected to be able to create DDD config file '%s' but it already exists." % args.filename
        )
    DddConfig().write_default_config(args.filename)


def create_rasa_config(args):
    if os.path.exists(args.filename):
        raise ConfigAlreadyExistsException(
            "Expected to be able to create RASA config file '%s' but it already exists." % args.filename
        )
    RasaConfig().write_default_config(args.filename)


def verify(args):
    backend_dependencies = BackendDependenciesForGenerating(args)

    ddds_builder = DDDBuilderForGenerating(
        backend_dependencies,
        verbose=args.verbose,
        ignore_warnings=args.ignore_warnings,
        language_codes=args.language_codes
    )
    ddds_builder.verify()


def _check_ddds_for_word_lists(ddds):
    for ddd in ddds:
        ddd_maker_utils.potentially_create_word_list_boilerplate(ddd.name)


@contextlib.contextmanager
def _config_exception_handling():
    def generate_message(name, command_name, config):
        return "Expected {name} config '{config}' to exist but it was not found. To create it, " "run 'tala {command} --filename {config}'.".format(
            name=name, command=command_name, config=config
        )

    try:
        yield
    except BackendConfigNotFoundException as exception:
        message = generate_message("backend", "create-backend-config", exception.config_path)
        raise ConfigNotFoundException(message)
    except DddConfigNotFoundException as exception:
        message = generate_message("DDD", "create-ddd-config", exception.config_path)
        raise ConfigNotFoundException(message)
    except RasaConfigNotFoundException as exception:
        message = generate_message("RASA", "create-rasa-config", exception.config_path)
        raise ConfigNotFoundException(message)


def version(args):
    print installed_version.version


def add_verify_subparser(subparsers):
    parser = subparsers.add_parser(
        "verify", help="verify the format of all DDDs supported by the backend, across all supported languages"
    )
    parser.set_defaults(func=verify)
    add_common_backend_arguments(parser)
    parser.add_argument(
        "--languages",
        dest="language_codes",
        choices=languages.SUPPORTED_LANGUAGES,
        nargs="*",
        help="override the backend config and verify the DDDs for these languages"
    )
    parser.add_argument("-v", "--verbose", action="store_true", dest="verbose")
    parser.add_argument(
        "--ignore-grammar-warnings",
        dest="ignore_warnings",
        action="store_true",
        help="ignore warnings when compiling the grammar"
    )
    parser.add_argument(
        "--rasa-config",
        dest="rasa_config",
        default=None,
        help="override the default RASA config %r" % RasaConfig.default_name()
    )


def add_create_ddd_subparser(subparsers):
    parser = subparsers.add_parser("create-ddd", help="create a new DDD")
    parser.set_defaults(func=create_ddd)
    parser.add_argument("name", help="Name of the DDD, e.g. my_ddd")
    parser.add_argument(
        "--target-dir", default=".", metavar="DIR", help="target directory, will be created if it doesn't exist"
    )


def add_create_backend_config_subparser(subparsers):
    parser = subparsers.add_parser("create-backend-config", help="create a backend config")
    parser.set_defaults(func=create_backend_config)
    parser.add_argument(
        "--filename",
        default=BackendConfig.default_name(),
        metavar="NAME",
        help="filename of the backend config, e.g. %s" % BackendConfig.default_name()
    )
    parser.add_argument("--ddd", help="name of the active DDD, e.g. my_ddd")


def add_create_ddd_config_subparser(subparsers):
    parser = subparsers.add_parser("create-ddd-config", help="create a DDD config")
    parser.set_defaults(func=create_ddd_config)
    parser.add_argument(
        "--filename",
        default=DddConfig.default_name(),
        metavar="NAME",
        help="filename of the DDD config, e.g. %s" % DddConfig.default_name()
    )


def add_create_rasa_config_subparser(subparsers):
    parser = subparsers.add_parser("create-rasa-config", help="create a TDM-specific RASA config")
    parser.set_defaults(func=create_rasa_config)
    parser.add_argument(
        "--filename",
        default=RasaConfig.default_name(),
        metavar="NAME",
        help="filename of the RASA config, e.g. %s" % RasaConfig.default_name()
    )


def add_version_subparser(subparsers):
    parser = subparsers.add_parser("version", help="print the Tala version")
    parser.set_defaults(func=version)


def format_warnings():
    def warning_on_one_line(message, category, _filename, _lineno, _file=None, _line=None):
        string = "%s: %s\n" % (category.__name__, message)
        return console_formatting.bold(string)

    warnings.formatwarning = warning_on_one_line


def show_deprecation_warnings():
    warnings.simplefilter("always", category=DeprecationWarning)


def main(args=None):
    format_warnings()
    show_deprecation_warnings()
    root_parser = argparse.ArgumentParser(description="Use the Tala SDK for the Talkamatic Dialogue Manager (TDM).")
    subparsers = root_parser.add_subparsers()
    add_verify_subparser(subparsers)
    add_create_ddd_subparser(subparsers)
    add_create_backend_config_subparser(subparsers)
    add_create_ddd_config_subparser(subparsers)
    add_create_rasa_config_subparser(subparsers)
    add_version_subparser(subparsers)

    parsed_args = root_parser.parse_args(args)
    with _config_exception_handling():
        parsed_args.func(parsed_args)


if __name__ == "__main__":
    main()



import re
import warnings


class DddDevice(object):
    """This method is invoked when a commitment is added or revised, e.g. when a user answer
    is integrated. The device can use this information to update its state.

    Return value: None."""

    def commitment_update(self, commitment):
        pass


class DeviceError(Exception):
    pass


class ParameterNotFoundException(Exception):
    pass


class ParameterField:
    VALUE = "value"
    GRAMMAR_ENTRY = "grammar_entry"


class DeviceMethod(object):
    def __init__(self, device, ontology):
        super(DeviceMethod, self).__init__()
        self.device = device
        self.ontology = ontology
        self.parameters = self._parse_parameters()
        self._check_parameters_sanity()
        self._current_invocation_id = None

    def _parse_parameters(self):
        if hasattr(self, "PARAMETERS"):
            return map(self._parse_parameter, self.PARAMETERS)
        return []

    def _parse_parameter(self, parameters_string):
        m = re.search("^([^=.]+)(\.(value|grammar_entry))?(='([^']*)')?$", parameters_string)
        if m:
            name, _, field, default_value_string, default_value = m.groups()

            if field is None:
                field = ParameterField.VALUE

            if default_value_string:
                optional = True
            else:
                optional = False
                default_value = None

            return {"name": name, "field": field, "optional": optional, "default_value": default_value}
        else:
            raise DeviceError(
                "failed to parse parameters for %s: invalid syntax %r" % (self.__class__.__name__, parameters_string)
            )

    def _check_parameters_sanity(self):
        for parameter in self.parameters:
            self._check_parameter_sanity(parameter)

    def _check_parameter_sanity(self, parameter):
        if not self.ontology.has_predicate(parameter["name"]):
            raise DeviceError(
                "illegal parameters for %s: no predicate %r found in the ontology %s" %
                (self.__class__.__name__, parameter["name"], self.ontology)
            )

    def current_invocation_id(self):
        return self._current_invocation_id

    def set_current_invocation_id(self, invocation_id):
        self._current_invocation_id = invocation_id


class EntityRecognizer(DeviceMethod):
    def recognize(self, string, language):
        if hasattr(self, "recognize_entity"):
            deprecation_message = "The 'recognize_entity' method does not distinguish between languages. " "Use the 'recognize' method instead. This happened in '%s' of '%s'." % (
                self.__class__.__name__, self.device.__class__.__name__
            )
            warnings.warn(deprecation_message, DeprecationWarning)
            return self.recognize_entity(string)
        raise NotImplementedError("This method must be implemented by a subclass if used.")

    def recognize_entity(self, string):
        pass


class DeviceAction(DeviceMethod):
    TYPE = None
    PLAY_AUDIO = "PLAY_AUDIO"


class DeviceWHQuery(DeviceMethod):
    pass


class Validity(DeviceMethod):
    pass


class DeviceImplication(DeviceMethod):
    def __init__(self, device, ontology):
        super(DeviceImplication, self).__init__(device, ontology)
        warnings.warn("%s methods are experimental and should be used with care" % DeviceImplication.__name__)



import os

from tala.ddd.building.steps.step_factory import CleanStepFactory, GenerateStepFactory, VerifyStepFactory
from tala.ddd.building.supported_asrs import SUPPORTED_ASRS
from tala.utils import chdir


class DDDsNotSpecifiedException(Exception):
    pass


class DDDNotFoundException(Exception):
    pass


class UnexpectedASRException(Exception):
    pass


class DDDBuilderForGenerating(object):
    def __init__(self, backend_dependencies, verbose=False, ignore_warnings=False, language_codes=None):
        super(DDDBuilderForGenerating, self).__init__()
        self._verbose = verbose
        self._ignore_warnings = ignore_warnings
        self._backend_dependencies = backend_dependencies
        self._language_codes = language_codes or self._backend_dependencies.supported_languages

        if self._backend_dependencies.asr not in SUPPORTED_ASRS:
            raise UnexpectedASRException(
                "Expected ASR as one of %s but got %r" % (SUPPORTED_ASRS, self._backend_dependencies.asr)
            )

        if not self._backend_dependencies.ddds:
            raise DDDsNotSpecifiedException("Expected DDDs to be specified in the backend config, but found none.")

        self.cwd = os.getcwd()
        self._cleaners = {}
        self._generators = {}
        self._verifiers = {}

        for ddd in self._backend_dependencies.ddds:
            if not os.path.exists(ddd.name):
                absolute_ddd_path = os.path.join(os.getcwd(), ddd.name)
                raise DDDNotFoundException("Expected a DDD at %r but found none." % absolute_ddd_path)
        for ddd in self._backend_dependencies.ddds:
            cleaner_factory = CleanStepFactory(ddd, self._language_codes, self._verbose, self._ignore_warnings)
            self._cleaners[ddd.name] = cleaner_factory.create()
            generator_factory = GenerateStepFactory(ddd, self._language_codes, self._verbose, self._ignore_warnings)
            self._generators[ddd.name] = generator_factory.create()
            verifier_factory = VerifyStepFactory(ddd, self._language_codes, self._verbose, self._ignore_warnings)
            self._verifiers[ddd.name] = verifier_factory.create()

    def build(self):
        for ddd in self._backend_dependencies.ddds:
            with chdir.chdir(ddd.name):
                print("Generating models for DDD '{}'.".format(ddd.name))
                self._cleaners[ddd.name].build()
                self._generators[ddd.name].build()
                print("Finished generating models for DDD '{}'.".format(ddd.name))

    def verify(self):
        for ddd_name, verifier in self._verifiers.items():
            with chdir.chdir(ddd_name):
                verifier.build()



from pathlib import Path

from tala.ddd.building.steps.clean import CleanStepForGFGeneration, CleanStepForHandcraftedGFFiles
from tala.ddd.building.steps.generate import GenerateStepForGFGeneration, GenerateStepForGFRGLGeneration, GenerateStepForHandcraftedGFFiles
from tala.ddd.building.steps.verify import VerifyStepForGFGeneration, VerifyStepForGFRGLGeneration, VerifyStepForHandcraftedGFFiles
from tala.ddd import utils


class AbstractStepFactory(object):
    def __init__(self, ddd, language_codes, verbose, ignore_warnings):
        self._ddd = ddd
        self._language_codes = language_codes
        self._verbose = verbose
        self._ignore_warnings = ignore_warnings

        self._grammar_directory = "grammar"
        grammar_path = Path.cwd() / self._ddd.name / self._grammar_directory
        self._has_handcrafted_gf_grammar = any(
            utils.has_handcrafted_gf_grammar(self._ddd.name, lang, str(grammar_path)) for lang in self._language_codes
        )

        self._ddd_root_directory = Path.cwd() / self._ddd.name

    def _create_step_for_gf_generation(self, *args, **kwargs):
        raise NotImplementedError()

    def _create_step_for_gf_rgl_generation(self, *args, **kwargs):
        raise NotImplementedError()

    def _create_step_for_handcrafted_gf_files(self, *args, **kwargs):
        raise NotImplementedError()

    def create(self):
        if self._has_handcrafted_gf_grammar:
            return self._create_step_for_handcrafted_gf_files()
        if self._ddd.use_rgl:
            return self._create_step_for_gf_rgl_generation()
        return self._create_step_for_gf_generation()

    def _create(self, class_):
        return class_(
            self._ddd, self._ignore_warnings, self._language_codes, self._verbose, str(self._ddd_root_directory),
            self._grammar_directory
        )


class CleanStepFactory(AbstractStepFactory):
    def _create_step_for_gf_generation(self):
        return self._create(CleanStepForGFGeneration)

    def _create_step_for_gf_rgl_generation(self):
        return self._create(CleanStepForGFGeneration)

    def _create_step_for_handcrafted_gf_files(self):
        return self._create(CleanStepForHandcraftedGFFiles)


class GenerateStepFactory(AbstractStepFactory):
    def _create_step_for_gf_generation(self):
        return self._create(GenerateStepForGFGeneration)

    def _create_step_for_gf_rgl_generation(self):
        return self._create(GenerateStepForGFRGLGeneration)

    def _create_step_for_handcrafted_gf_files(self):
        return self._create(GenerateStepForHandcraftedGFFiles)


class VerifyStepFactory(AbstractStepFactory):
    def _create_step_for_gf_generation(self):
        return self._create(VerifyStepForGFGeneration)

    def _create_step_for_gf_rgl_generation(self):
        return self._create(VerifyStepForGFRGLGeneration)

    def _create_step_for_handcrafted_gf_files(self):
        return self._create(VerifyStepForHandcraftedGFFiles)



from tala.model.semantic_object import OntologySpecificSemanticObject, SemanticObject


class LambdaAbstractedPredicateProposition(OntologySpecificSemanticObject):
    def __init__(self, predicate, ontology_name):
        OntologySpecificSemanticObject.__init__(self, ontology_name)
        self.predicate = predicate

    def is_lambda_abstracted_predicate_proposition(self):
        return True

    def __unicode__(self):
        variable = "X"
        return variable + "." + self.getPredicate().get_name() + "(" + variable + ")"

    def __eq__(self, other):
        if (isinstance(other, LambdaAbstractedPredicateProposition)):
            return self.getPredicate() == other.getPredicate()
        else:
            return False

    def __ne__(self, other):
        return not (self == other)

    def getPredicate(self):
        return self.predicate

    def getSort(self):
        return self.predicate.getSort()

    def __hash__(self):
        return hash((self.__class__.__name__, self.predicate))


class LambdaAbstractedGoalProposition(SemanticObject):
    def __init__(self):
        SemanticObject.__init__(self)

    def is_lambda_abstracted_goal_proposition(self):
        return True

    def __eq__(self, other):
        return isinstance(other, LambdaAbstractedGoalProposition)

    def __ne__(self, other):
        return not (self == other)

    def __unicode__(self):
        return "X.goal(X)"

    def __hash__(self):
        return hash(self.__class__.__name__)



from tala.model.semantic_object import OntologySpecificSemanticObject


class Predicate(OntologySpecificSemanticObject):
    def __init__(self, ontology_name, name, sort, feature_of_name=None, multiple_instances=False):
        OntologySpecificSemanticObject.__init__(self, ontology_name)
        self.name = name
        self.sort = sort
        self.feature_of_name = feature_of_name
        self._multiple_instances = multiple_instances

    def get_name(self):
        return self.name

    def getSort(self):
        return self.sort

    def get_feature_of_name(self):
        return self.feature_of_name

    def is_feature_of(self, other):
        return self.feature_of_name == other.get_name()

    def allows_multiple_instances(self):
        return self._multiple_instances

    def is_predicate(self):
        return True

    def __eq__(self, other):
        try:
            return isinstance(other, Predicate) and other.get_name() == self.get_name() and other.getSort(
            ) == self.getSort() and other.get_feature_of_name() == self.get_feature_of_name(
            ) and other.allows_multiple_instances() == self.allows_multiple_instances()
        except AttributeError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash((self.name, self.sort, self.feature_of_name, self._multiple_instances))

    def __str__(self):
        return self.get_name()

    def __repr__(self):
        return "%s%s" % (
            self.__class__.__name__, (self.name, self.sort, self.feature_of_name, self._multiple_instances)
        )



from tala.model.action import Action
from tala.model.error import DomainError
from tala.model.goal import PerformGoal, ResolveGoal
from tala.model.speaker import Speaker
from tala.model.plan import Plan, InvalidPlansException
from tala.model.proposition import PredicateProposition, ServiceActionTerminatedProposition
from tala.model.question_raising_plan_item import QuestionRaisingPlanItem
from tala.utils.as_json import JSONLoggable


class DddDomain:
    pass


class Domain(JSONLoggable):
    DEFAULT_IO_STATUS = "default"
    EXCLUDED_IO_STATUS = "excluded"
    HIDDEN_IO_STATUS = "hidden"
    SILENT_IO_STATUS = "silent"
    DISABLED_IO_STATUS = "disabled"

    def __init__(self, ddd_name, name, ontology, plans=[], default_questions=[], parameters={}, dependencies={}):
        self.ddd_name = ddd_name
        self.name = name
        self.ontology = ontology
        self.default_questions = default_questions
        self.parameters = parameters
        self.dependencies = dependencies
        self.plans = self._plan_list_to_dict_indexed_by_goal(plans)
        self._goals_in_defined_order = [plan["goal"] for plan in plans]
        self._add_top_plan_if_missing()
        self._add_up_plan()

    def as_json(self):
        json = super(Domain, self).as_json()
        json["ontology"] = "<skipped>"
        return json

    @property
    def goals(self):
        return self.plans.keys()

    def _has_top_plan(self):
        return any(filter(lambda plan:
                              (plan["goal"].is_goal() and
                               plan["goal"].get_goal_type() == PerformGoal.PERFORM_GOAL and
                               plan["goal"].get_action().get_value() == "top"),
                          self.plans.values()))  # noqa: E127

    def get_ontology(self):
        return self.ontology

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return self.name == other.name and self.ddd_name == other.ddd_name

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.name) * hash(self.ddd_name)

    def get_name(self):
        return self.name

    def action(self, action_name):
        return Action(action_name, self.ontology.get_name())

    def is_depending_on(self, dependent_issue, other):
        if dependent_issue in self.dependencies:
            if other in self.dependencies[dependent_issue]:
                return True
        if self._goal_issue_depends_on_question(dependent_issue, other):
            return True
        if self.is_feature_of(other, dependent_issue):
            return True
        return False

    def is_feature_of(self, feature_question, question):
        try:
            feature_question_predicate = feature_question.get_content().getPredicate()
            question_predicate = question.get_content().getPredicate()
        except AttributeError:
            return False
        return feature_question_predicate.is_feature_of(question_predicate)

    def _goal_issue_depends_on_question(self, goal_issue, question):
        resolve_goal = ResolveGoal(goal_issue, Speaker.SYS)
        if self.has_goal(resolve_goal):
            plan = self.get_plan(resolve_goal)
            for item in plan:
                if item.is_question_plan_item():
                    if question == item.getContent():
                        return True
        return False

    def get_dependent_question(self, question):
        for other_question in self.get_plan_questions():
            if self.is_depending_on(other_question, question):
                return other_question
        return None

    def has_goal(self, goal):
        return goal in self.plans

    def get_plan(self, goal):
        plan_info = self._get_plan_info(goal)
        return plan_info["plan"]

    def _get_plan_info(self, goal):
        if goal in self.plans:
            return self.plans[goal]
        else:
            raise DomainError("no plan for goal '%s' in domain '%s'" % (goal, self.get_name()))

    def get_io_status(self, goal):
        if self.has_goal(goal):
            plan_info = self._get_plan_info(goal)
            if "io_status" in plan_info:
                return plan_info["io_status"]
        return Domain.DEFAULT_IO_STATUS

    def dominates(self, supergoal, subgoal):
        if self.has_goal(supergoal):
            plan = self.get_plan(supergoal)
            for item in plan:
                if item.is_question_plan_item():
                    question = item.getContent()
                    if question.is_alt_question():
                        for proposition in question.get_content():
                            if proposition.is_goal_proposition():
                                goal = proposition.get_goal()
                            if goal == subgoal:
                                return True
                            elif self.dominates(goal, subgoal):
                                return True
        return False

    def _is_action_goal(self, goal):
        try:
            return goal.is_action()
        except AttributeError:
            return False

    def goal_is_preferred(self, goal):
        return self.goal_is_conditionally_preferred(goal, [])

    def goal_is_conditionally_preferred(self, goal, facts):
        condition = self.get_preferred(goal)
        if condition is True or condition in facts:
            return True

    def is_default_question(self, question):
        return question in self.default_questions

    def get_plan_questions(self):
        already_found = set()
        for goal in self.plans:
            plan = self.get_plan(goal)
            for question in self.get_questions_in_plan(plan):
                if question not in already_found:
                    already_found.add(question)
                    yield question

    def get_plan_goal_iterator(self):
        for goal in self.plans:
            yield goal

    def get_all_goals(self):
        return self.plans.keys()

    def get_all_goals_in_defined_order(self):
        return self._goals_in_defined_order

    def get_all_resolve_goals(self):
        return filter(lambda x: x.is_resolve_goal(), self.plans)

    def get_postconds(self, goal):
        if self.has_goal(goal):
            for postcond in self.get_goal_attribute(goal, 'postconds'):
                yield postcond
            for postcond in self._implicit_postconds_for_service_action_invocations(goal):
                yield postcond
            for postcond in self._implicit_postconds_for_handle_goal(goal):
                yield postcond

    def _implicit_postconds_for_service_action_invocations(self, goal):
        plan = self.get_plan(goal)
        for item in plan:
            if item.is_invoke_service_action_plan_item() and item.should_downdate_plan():
                yield ServiceActionTerminatedProposition(self.ontology.name, item.get_service_action())

    def _implicit_postconds_for_handle_goal(self, goal):
        if goal.is_handle_goal():
            yield ServiceActionTerminatedProposition(self.ontology.name, goal.get_device_event())

    def get_goal_attribute(self, goal, attribute):
        plan_info = self._get_plan_info(goal)
        try:
            return plan_info[attribute]
        except KeyError:
            return []

    def get_postplan(self, goal):
        if self.has_goal(goal):
            return self.get_goal_attribute(goal, 'postplan')
        else:
            return []

    def get_superactions(self, goal):
        if self.has_goal(goal):
            return self.get_goal_attribute(goal, 'superactions')
        else:
            return []

    def get_gui_context(self, goal):
        return self.get_goal_attribute(goal, 'gui_context')

    def restart_on_completion(self, goal):
        return self.get_goal_attribute(goal, 'restart_on_completion')

    def has_dynamic_title(self, goal):
        return self.get_goal_attribute(goal, 'dynamic_title')

    def get_reraise_on_resume(self, goal):
        reraise_on_resume = self.get_goal_attribute(goal, "reraise_on_resume")
        if reraise_on_resume == []:
            reraise_on_resume = True
        return reraise_on_resume

    def is_silently_accommodatable(self, goal):
        return self.get_goal_attribute(goal, 'unrestricted_accommodation')

    def get_preferred(self, goal):
        if self.has_goal(goal):
            return self.get_goal_attribute(goal, 'preferred')
        else:
            return []

    def goal_allows_accommodation_without_feedback(self, goal):
        return self.get_goal_attribute(goal, "accommodate_without_feedback")

    def get_resolving_answers(self, question):
        predicate = question.get_predicate()
        question_sort = predicate.getSort()
        answers = set()
        for individual_value in self.ontology.get_individuals():
            individual_sort = self.ontology.individual_sort(individual_value)
            if individual_sort == question_sort:
                individual = self.ontology.create_individual(individual_value)
                answer = PredicateProposition(predicate, individual)
                answers.add(answer)
        return answers

    def _plan_list_to_dict_indexed_by_goal(self, plan_list):
        plans = {}
        for plan_info in plan_list:
            goal = plan_info["goal"]
            if not goal.is_goal():
                raise Exception("expected goal but found %s" % goal)
            if goal in plans:
                raise InvalidPlansException("multiple plans for goal %s" % goal)
            plans[goal] = plan_info
        return plans

    def _add_up_plan(self):
        up_action = self.action("up")
        goal = PerformGoal(up_action)
        self.plans[goal] = {"goal": goal, "plan": Plan()}

    def _add_top_plan_if_missing(self):
        top_goal = PerformGoal(self.action("top"))
        if top_goal not in self.plans:
            self.plans[top_goal] = {"goal": top_goal, "plan": Plan()}

    def get_graphical_type(self, semantic_object):
        return self._get_parameter(semantic_object, "graphical_type")

    def get_incremental(self, semantic_object):
        return self._get_parameter(semantic_object, "incremental")

    def get_alternatives(self, semantic_object):
        return self._get_parameter(semantic_object, "alts")

    def get_source(self, semantic_object):
        return self._get_parameter(semantic_object, "source")

    def get_format(self, semantic_object):
        return self._get_parameter(semantic_object, "format")

    def get_default(self, semantic_object):
        return self._get_parameter(semantic_object, "default")

    def get_service_query(self, semantic_object):
        return self._get_parameter(semantic_object, "service_query")

    def get_label_questions(self, semantic_object):
        return self._get_parameter(semantic_object, "label_questions")

    def get_sort_order(self, semantic_object):
        return self._get_parameter(semantic_object, "sort_order")

    def get_max_spoken_alts(self, question):
        return self._get_parameter(question, "max_spoken_alts")

    def get_background(self, semantic_object):
        return self._get_parameter(semantic_object, "background")

    def get_ask_features(self, semantic_object):
        return self._get_parameter(semantic_object, "ask_features")

    def get_related_information(self, semantic_object):
        return self._get_parameter(semantic_object, "related_information")

    def get_verbalize(self, semantic_object):
        if self._get_parameter(semantic_object, "verbalize") is False:
            return False
        else:
            return True

    def allows_goal_accommodation(self, question):
        if self._get_parameter(question, "allow_goal_accommodation") is False:
            return False
        else:
            return True

    def question_forces_graphical_choice(self, question):
        return self.get_graphical_type(question) in [
            QuestionRaisingPlanItem.GRAPHICAL_TYPE_LIST, QuestionRaisingPlanItem.GRAPHICAL_TYPE_TEXT
        ] and not self.get_incremental(question)

    def _get_parameter(self, question, parameter):
        try:
            return self.parameters[question][parameter]
        except KeyError:
            return None

    def get_questions_in_plan(self, plan):
        for plan_item in plan:
            if plan_item.is_question_plan_item():
                question = plan_item.getContent()
                for feature_question in self.get_feature_questions_for_plan_item(question, plan_item):
                    yield feature_question
                yield question

    def get_feature_questions_for_plan_item(self, question, plan_item):
        feature_questions = []
        if question.get_content().is_lambda_abstracted_predicate_proposition():
            for predicate in self.ontology.get_predicates().values():
                if predicate.is_feature_of(question.get_content().getPredicate()):
                    feature_question = self.ontology.create_wh_question(predicate.get_name())
                    feature_questions.append(feature_question)
        return feature_questions

    def get_invoke_service_action_items_for_action(self, action_name):
        for goal in self.get_all_goals():
            plan = self.get_plan(goal)
            for item in plan:
                if item.is_invoke_service_action_plan_item() and item.get_service_action() == action_name:
                    yield item



from tala.config import BackendConfig

INACTIVE_SECONDS_ALLOWED_MIN = 60 * 60
INACTIVE_SECONDS_ALLOWED_MAX = 60 * 60 * 24 * 30


class InvalidConfigValue(Exception):
    pass


class AbstractBackendDependencies(object):
    def __init__(self, args):
        super(AbstractBackendDependencies, self).__init__()

        self._overridden_ddd_config_paths = args.overridden_ddd_config_paths
        if hasattr(args, "interpret_partially"):
            self._interpret_partially = args.interpret_partially
        else:
            self._interpret_partially = None
        self._raw_config = BackendConfig(args.config).read()

        self._asr = self._raw_config["asr"]
        self._use_recognition_profile = self._raw_config["use_recognition_profile"]
        self._repeat_questions = self._raw_config["repeat_questions"]
        self._use_word_list_correction = self._raw_config["use_word_list_correction"]
        self._supported_languages = self._raw_config["supported_languages"]
        self._ddd_names = self._raw_config["ddds"]
        self._rerank_amount = self._raw_config["rerank_amount"]
        self._inactive_seconds_allowed = self._raw_config["inactive_seconds_allowed"]
        self._validate_inactive_seconds_allowed()

        self.ddds = self.load_ddds(self._raw_config["ddds"])

    def _validate_inactive_seconds_allowed(self):
        if self._inactive_seconds_allowed < INACTIVE_SECONDS_ALLOWED_MIN or self._inactive_seconds_allowed > INACTIVE_SECONDS_ALLOWED_MAX:
            raise InvalidConfigValue(
                "Expected inactive_seconds_allowed to be in the range %d-%d, but it was %d." %
                (INACTIVE_SECONDS_ALLOWED_MIN, INACTIVE_SECONDS_ALLOWED_MAX, self._inactive_seconds_allowed)
            )

    def load_ddds(self, ddd_names):
        raise NotImplementedError("Expected method _load_ddds(...) to be implemented but it wasn't")

    @property
    def raw_config(self):
        return self._raw_config

    @property
    def overridden_ddd_config_paths(self):
        return self._overridden_ddd_config_paths

    @property
    def interpret_partially(self):
        return self._interpret_partially

    @property
    def ddds(self):
        return self._ddds

    @ddds.setter
    def ddds(self, ddds):
        self._ddds = ddds

    @property
    def asr(self):
        return self._asr

    @property
    def use_recognition_profile(self):
        return self._use_recognition_profile

    @property
    def repeat_questions(self):
        return self._repeat_questions

    @property
    def use_word_list_correction(self):
        return self._use_word_list_correction

    @property
    def supported_languages(self):
        return self._supported_languages

    @property
    def ddd_names(self):
        return self._ddd_names

    @property
    def rerank_amount(self):
        return self._rerank_amount

    @property
    def inactive_seconds_allowed(self):
        return self._inactive_seconds_allowed

    @property
    def ddds_with_rasa(self):
        return [ddd for ddd in self.ddds if ddd.is_rasa_enabled]



import os
from copy import copy
import json

from jinja2 import Template
from pathlib import Path

from tala.ddd.grammar.reader import GrammarReader
from tala.nl.rasa.constants import ACTION_INTENT, QUESTION_INTENT, ANSWER_INTENT, ANSWER_NEGATION_INTENT
from tala.nl.rasa.generating.common_example import CommonExample
from tala.nl.rasa.generating.entity_factory import PropositionalEntityFactory, SortalEntityFactory
from tala.nl.rasa.generating.examples import Examples
from tala.utils.file_writer import UTF8FileWriter


class RASANotSupportedByGrammarFormatException(Exception):
    pass


class UnexpectedRequiredEntityException(Exception):
    pass


class RASADataNotGeneratedException(Exception):
    pass


class RasaGenerator(object):
    def __init__(self, ddd, language_code):
        super(RasaGenerator, self).__init__()
        self._ddd = ddd
        self._sortal_entity_factory = SortalEntityFactory()
        self._propositional_entity_factory = PropositionalEntityFactory()
        self._language_code = language_code
        self._language_examples = Examples.from_language(language_code)

    def generate(self):
        if not GrammarReader.xml_grammar_exists_for_language(self._language_code):
            raise RASANotSupportedByGrammarFormatException(
                "Expected an XML-based grammar at '%s', but it does not exist" %
                os.path.abspath(GrammarReader.path(self._language_code))
            )
        grammar = self._ddd.grammars[self._language_code]

        examples = set()
        examples.update(self._examples_of_requests(grammar))
        examples.update(self._examples_of_questions(grammar))
        examples.update(self._examples_of_sortal_answers_from_individuals(grammar))
        examples.update(self._examples_of_sortal_answer_negations_from_individuals(grammar))
        examples.update(self._examples_of_answers(grammar))
        examples.update(self._examples_of_negative_intent())

        rasa_data = {
            "rasa_nlu_data": {
                "regex_features": [],
                "entity_synonyms": [],
                "common_examples": [example.generate() for example in examples]
            }
        }

        return rasa_data

    @staticmethod
    def data_file_name():
        return "rasa_data.json"

    def generate_and_write_to_file(self):
        def write(path, text):
            writer = UTF8FileWriter(path)
            writer.create_directories()
            writer.write(text)

        def as_json(data):
            return json.dumps(data, ensure_ascii=False, indent=4, separators=(',', ': '))

        path = Path("build_rasa") / self._language_code / self.data_file_name()
        data = self.generate()
        write(path, as_json(data))

    def _examples_of_negative_intent(self):
        return self._language_examples.negative

    def _examples_of_requests(self, grammar):
        for action in self._ddd.ontology.get_ddd_specific_actions():
            for example in self._examples_of_request(grammar, action):
                yield example

    def _examples_of_request(self, grammar, action):
        intent = "%s:%s::%s" % (self._ddd.name, ACTION_INTENT, action)
        requests = grammar.requests_of_action(action)
        for request in requests:
            for example in self._examples_of_intent(grammar, intent, request):
                yield example

    def _examples_of_intent(self, grammar, name, intent):
        head = intent.text_chunks[0]
        texts = intent.text_chunks[1:]
        first_example = CommonExample(name, head)
        examples = self._examples_with_individuals(grammar, texts, intent.required_entities, [first_example])
        for example in examples:
            yield example

    def _examples_with_individuals(self, grammar, text_chunks, required_entities, examples_so_far):
        if not text_chunks and not required_entities:
            return examples_so_far
        tail = text_chunks[0]
        required_entity = required_entities[0]
        all_new_examples = []
        for example in examples_so_far:
            if required_entity.is_sortal:
                new_examples = list(self._examples_from_sortal_individual(grammar, required_entity, example, tail))
            elif required_entity.is_propositional:
                new_examples = list(
                    self._examples_from_propositional_individual(grammar, required_entity, example, tail)
                )
            else:
                raise UnexpectedRequiredEntityException(
                    "Expected either a sortal or propositional required entity but got a %s" %
                    required_entity.__class__.__name__
                )
            all_new_examples.extend(new_examples)
        return self._examples_with_individuals(grammar, text_chunks[1:], required_entities[1:], all_new_examples)

    def _examples_from_sortal_individual(self, grammar, required_sortal_entity, example_so_far, tail):
        sort = self._ddd.ontology.get_sort(required_sortal_entity.name)
        individuals = self._individual_grammar_entries_samples(grammar, sort)
        return self._examples_from_individuals(
            self._sortal_entity_factory, sort.get_name(), individuals, example_so_far, tail
        )

    def _sample_individuals_of_builtin_sort(self, sort):
        examples = self._language_examples.get_builtin_sort_examples(sort)
        return [[entry] for entry in examples]

    def _individual_grammar_entries_samples(self, grammar, sort):
        if sort.is_builtin():
            return self._sample_individuals_of_builtin_sort(sort)
        return self._individual_grammar_entries_samples_of_custom_sort(grammar, sort)

    def _individual_grammar_entries_samples_of_custom_sort(self, grammar, sort):
        individuals = self._ddd.ontology.get_individuals_of_sort(sort.get_name())
        grammar_entries = [grammar.entries_of_individual(individual) for individual in individuals]
        return grammar_entries

    def _examples_from_propositional_individual(self, grammar, required_propositional_entity, example_so_far, tail):
        predicate_name = required_propositional_entity.name
        predicate = self._ddd.ontology.get_predicate(predicate_name)
        sort = predicate.getSort()
        individuals = self._individual_grammar_entries_samples(grammar, sort)
        if sort.is_string_sort():
            predicate_specific_samples = self._string_examples_of_predicate(grammar, predicate)
            individuals.extend([[predicate_specific_sample]
                                for predicate_specific_sample in predicate_specific_samples])
        return self._examples_from_individuals(
            self._propositional_entity_factory, predicate_name, individuals, example_so_far, tail
        )

    def _string_examples_of_predicate(self, grammar, predicate):
        return grammar.strings_of_predicate(predicate.get_name())

    def _examples_from_individuals(self, entity_factory, identifier, individuals_grammar_entries, example_so_far, tail):
        for grammar_entries in individuals_grammar_entries:
            for grammar_entry in grammar_entries:
                entity_start = len(example_so_far.text)
                end = entity_start + len(grammar_entry)
                entity = entity_factory.create(identifier, grammar_entry, entity_start, end)
                example = self._extend_example(entity, example_so_far, tail)
                yield example

    @staticmethod
    def _extend_example(entity, example_so_far, tail=None):
        head = example_so_far.text
        tail = tail or ""
        text = "".join([head, entity.text, tail])
        entities = copy(example_so_far.entities)
        entities.append(entity)
        return CommonExample(example_so_far.intent, text, entities)

    def _examples_of_questions(self, grammar):
        for resolve_goal in self._ddd.domain.get_all_resolve_goals():
            question = resolve_goal.get_question()
            for example in self._examples_of_question(grammar, question):
                yield example

    def _examples_of_question(self, grammar, question):
        predicate = question.get_predicate().get_name()
        intent = "%s:%s::%s" % (self._ddd.name, QUESTION_INTENT, predicate)
        questions = grammar.questions_of_predicate(predicate)
        for question in questions:
            for example in self._examples_of_intent(grammar, intent, question):
                yield example

    def _examples_of_answers(self, grammar):
        for answer in grammar.answers():
            intent = "%s:%s" % (self._ddd.name, ANSWER_INTENT)
            for example in self._examples_of_intent(grammar, intent, answer):
                yield example

    def _examples_of_sortal_answers_from_individuals(self, grammar):
        for sort in self._ddd.ontology.get_sorts().values():
            examples = self._examples_of_sortal_answers_of_kind(grammar, sort, ANSWER_INTENT, self._answer_templates)
            for example in examples:
                yield example

    def _examples_of_sortal_answers_of_kind(self, grammar, sort, kind, templates):
        for grammar_entries in self._individual_grammar_entries_samples(grammar, sort):
            examples = self._examples_of_individual(
                kind, self._sortal_entity_factory, grammar_entries, sort.get_name(), templates
            )
            for example in examples:
                yield example

    def _examples_of_sortal_answer_negations_from_individuals(self, grammar):
        for sort in self._ddd.ontology.get_sorts().values():
            if sort.is_string_sort():
                continue
            examples = self._examples_of_sortal_answers_of_kind(
                grammar, sort, ANSWER_NEGATION_INTENT, self._answer_negation_templates
            )
            for example in examples:
                yield example

    @property
    def _answer_templates(self):
        template = Template('{{ name }}')
        return [template]

    @property
    def _answer_negation_templates(self):
        template = Template('not {{ name }}')
        return [template]

    def _examples_of_individual(self, intent_type, entity_factory, grammar_entries, identifier, templates):
        intent = "%s:%s" % (self._ddd.name, intent_type)
        for grammar_entry in grammar_entries:
            examples = self._examples_from_templates(intent, entity_factory, grammar_entry, identifier, templates)
            for example in examples:
                yield example

    def _examples_from_templates(self, intent, entity_factory, grammar_entry, identifier, templates):
        for template in templates:
            text = template.render(name=grammar_entry)
            example = self._create_example_with_only_an_entity(intent, entity_factory, identifier, grammar_entry, text)
            yield example

    @staticmethod
    def _create_example_with_only_an_entity(intent, entity_factory, identifier, grammar_entry, text):
        start = text.find(grammar_entry)
        if start < 0:
            raise RuntimeError(
                "Expected to find individual named '%s' in example '%s' but it was not found." % (grammar_entry, text)
            )
        end = start + len(grammar_entry)
        entity = entity_factory.create(identifier, grammar_entry, start=start, end=end)
        return CommonExample(intent, text, [entity])



from collections import namedtuple

PROTOCOL_VERSION = "1.0"


class Session(namedtuple("Session", ["session_id"])):
    pass


class Context(namedtuple("Context", ["active_ddd", "facts", "language", "invocation_id"])):
    pass
